# -*- coding: utf-8 -*-
from enerpi.base import CONFIG, log, CONFIG_FILENAME


# PINOUT RGB_LED
WITH_RGBLED = CONFIG.getboolean('RGBLED', 'WITH_RGBLED', fallback=False)
PIN_R = CONFIG.getint('RGBLED', 'PIN_R', fallback=19)
PIN_G = CONFIG.getint('RGBLED', 'PIN_G', fallback=20)
PIN_B = CONFIG.getint('RGBLED', 'PIN_B', fallback=16)


def get_rgbled(verbose=True):
    """Tries to get gpiozero RGBLED object at pins PIN_R|G|B if WITH_RGBLED in config file"""
    led = None
    if WITH_RGBLED:
        try:
            from gpiozero import RGBLED
            from gpiozero.pins.rpigpio import RPiGPIOPin

            led = RGBLED(RPiGPIOPin(PIN_R), RPiGPIOPin(PIN_G), RPiGPIOPin(PIN_B), active_high=True)
            led.blink(.25, .25, on_color=(1, 1, 1), n=5)
        except (OSError, RuntimeError, ImportError) as e:
            log('** Not using RGBLED with GPIOZERO ({} [{}]). Check your "{}" file.'
                .format(e, e.__class__, CONFIG_FILENAME), 'warn', verbose)
    return led


def led_alarm(led, time_blinking=2.5, timeout=3, time_on=.25, color=(1, 0, 0)):
    """Blinks for alarm code (one time or multiple times, with ∆T=.25 sec or 'time_on')"""
    if timeout == 0:
        led.blink(time_on, time_on, on_color=color)
    else:
        led.blink(time_on, time_on, on_color=color, n=int(time_blinking / (2 * time_on)))


def led_info(led, n=3):
    """Blinks in BLUE n times with ∆T=.5 sec"""
    led.blink(.5, .5, on_color=(0, 0, 1), n=n)


def blink_color(led, color, n=1):
    """Blinks n times in color with ∆T=.5 sec"""
    led.blink(.5, .5, on_color=color, n=n)





# -*- coding: utf-8 -*-
import subprocess


def get_cpu_temp():
    """
    Read CPU Temperature from RPI

    :return: :float: Temp in ºC
    """
    cmd = "/sys/class/thermal/thermal_zone0/temp"
    try:
        with open(cmd, 'r') as temp_file:
            cpu_temp = temp_file.read()
        return float(cpu_temp) / 1000
    except FileNotFoundError:
        print('Is this a RPI? "{}" not found!'.format(cmd))
        return None


def get_gpu_temp():
    """
    Read GPU Temperature from RPI

    :return: :float: Temp in ºC
    """
    try:
        gpu_temp = subprocess.check_output(['/opt/vc/bin/vcgencmd', 'measure_temp']
                                           ).decode().replace("temp=", "").replace("'C", "")
        return float(gpu_temp)
    except FileNotFoundError:
        print('Is this a RPI? "{}" not found!'.format('/opt/vc/bin/vcgencmd'))
        return None



# -*- coding: utf-8 -*-
import datetime as dt
from gpiozero import MCP3008
import json
import numpy as np
import random
import re
from subprocess import check_output
from time import sleep, time
from enerpi.base import SENSORS, log


PREC_SAMPLING = .0002  # := dt.timedelta(microseconds=500)
MIN_NUM_SAMPLES_RMS = 10
HOST = check_output('hostname').decode().splitlines()[0]
EXAMPLE_POWER_EV = [(0, 300), (3, 1200), (6, 3250), (8, 700), (9, 100), (10, 0), (12, 300)]


class DummySensor(object):
    """
    Object for mimic a MCP Sensor.
    """
    def __init__(self, channel=0, power_noise=15,
                 power_divisor=SENSORS.rms_multiplier, power_evolution=EXAMPLE_POWER_EV,
                 signal_center=.49951124, freq_hz=50):
        self._ch = channel
        self._v_min = 0
        self._v_max = 1
        self._counter = 0
        self._center = signal_center
        self._power_divisor = power_divisor
        self._noise = power_noise / self._power_divisor
        times, values = list(zip(*power_evolution))
        self._true_value = values[-1] / self._power_divisor
        self._power_ev_time = np.array(times)
        self._power_ev_values = np.array(values)
        self._period_ev = self._power_ev_time[-1] + 1
        self._freq = freq_hz
        self._t0 = time()

    @property
    def elapsed(self):
        """Elapsed time from creation"""
        return time() - self._t0

    @property
    def is_active(self):
        """True, like GPIOZERO_MCP sensor.is_active"""
        return True

    @property
    def value(self):
        """Mimics reading of MCP sensor"""
        ts = self.elapsed % self._period_ev
        value = self._power_ev_values[np.where(self._power_ev_time < ts)[0][-1]]
        self._true_value = value / self._power_divisor
        noise_v = 2 * (random.random() - .5) * self._noise
        v = self._true_value + noise_v
        return self._center + np.sqrt(2) * v * np.sin((self._freq * 2 * np.pi) * self.elapsed)

    @property
    def closed(self):
        """Mimics closed state of MCP sensor"""
        return True


class AnalogSensorBuffer(object):
    """
    Wrapper for MCP Sensor with 1D ring buffer using numpy arrays for RMS & M sensing.
    """
    def __init__(self, probe, value_correction, length, rms_sensor=True):
        self._probe = probe
        self._value_correction = value_correction
        self.size_max = length
        self._rms_sensor = rms_sensor
        self.size = 0
        self._completed = False
        self._data = np.zeros(length, dtype='f')
        self._last_mean = 0.

    @property
    def software_spi(self):
        """True if MCP/GPIOZERO is using Software SPI (much slower than Hardware mode)"""
        str_repr = self._probe.__repr__()
        return 'using software SPI' in str_repr

    @property
    def is_active(self):
        """Returns GPIOZERO_MCP sensor.is_active"""
        return self._probe.is_active

    @property
    def is_rms(self):
        """True if it's a Root-Mean-Squared measuring GPIOZERO_MCP sensor"""
        return self._rms_sensor

    def append_reading(self):
        """Read MCP sensor values, and append with value_correction
        to obtain future Root-Mean-Squared (RMS) value or future simple Mean value"""
        x = self._probe.value + self._value_correction
        if self._rms_sensor:
            new = x**2
        else:
            new = x
        if not self._completed:
            self.size += 1
            self._last_mean += new / self.size_max
            if self.size == self.size_max:
                self._completed = True
        else:
            self._last_mean += (new - self._data[-1]) / self.size_max
        self._data = np.roll(self._data, 1)
        self._data[0] = new

    def mean(self):
        """Returns the mean of data in the ring buffer"""
        # if self.size == self.size_max:
        #     return np.mean(self._data)
        # else:
        #     return np.mean(self._data[:self.size])
        return self._last_mean

    def read(self):
        """Reads MCP sensor values, and appends value_correction"""
        return self._probe.value + self._value_correction

    def close_probe(self):
        """Closes MCP sensor"""
        if self._probe is not None and not self._probe.closed:
            self._probe.close()


def tuple_to_dict_json(data_tuple):
    """
    Makes JSON message for broadcasting, from tuple of sampling values.

    :param data_tuple: :tuple: sampling values
    :return: :str: JSON message
    """
    d_data = dict(zip(SENSORS.columns_sampling, data_tuple))
    d_data['host'] = HOST
    d_data[SENSORS.ts_column] = d_data[SENSORS.ts_column].strftime(SENSORS.ts_fmt)
    # d_data[SENSORS.ts_column] = dt.datetime.now().strftime(SENSORS.ts_fmt)

    cols_rms, cols_mean, cols_ref = SENSORS.included_columns_sampling(d_data)
    for c in cols_rms:
        d_data[c] = int(round(d_data[c]))
    for c in cols_mean:
        d_data[c] = round(d_data[c], 4)
    for c in cols_ref:
        d_data[c] = int(d_data[c])
    js = json.dumps(d_data)
    return js


def msg_to_dict(msg):
    """
    Makes dict of broadcasted message of sampling values.

    :param msg: :str: message
    :return: :dict: data dict
    """
    try:
        d_data = json.loads(msg)
    except json.decoder.JSONDecodeError:
        # Versión anterior:
        rg_msg_mask = re.compile('^(?P<host>.*) __ (?P<ts>.*) __ (?P<power>.*) W __ Noise: (?P<noise>.*) W __ '
                                 'REF: (?P<ref>.*) __ LDR: (?P<ldr>.*)')
        d_data = rg_msg_mask.search(msg).groupdict()
        for k in filter(lambda x: x in d_data, SENSORS.columns_sensors_rms + SENSORS.columns_sensors_mean):
            d_data[k] = float(d_data[k])
    d_data[SENSORS.ts_column] = SENSORS.TZ.localize(dt.datetime.strptime(d_data[SENSORS.ts_column], SENSORS.ts_fmt))
    d_data['msg'] = msg
    return d_data


def _sampler(n_samples_buffer=SENSORS.n_samples_buffer_rms, delta_sampling=SENSORS.delta_sec_data,
             min_ts_ms=SENSORS.ts_data_ms, delta_secs_raw_capture=None,
             measure_ldr_divisor=SENSORS.measure_ldr_divisor,
             use_dummy_sensors=False, reset_bias=False, verbose=False):
    delta_sampling_calc = delta_sampling
    con_pausa = min_ts_ms > 0
    buffers = []
    assert(len(SENSORS) > 0)
    if reset_bias:
        for s in SENSORS:
            s.bias = 0
    n_samples_buffer_rms = n_samples_buffer
    n_samples_buffer_normal = n_samples_buffer // measure_ldr_divisor
    if use_dummy_sensors:
        sensor_probe, probe_type = DummySensor, 'DummySensor'
    else:
        sensor_probe, probe_type = MCP3008, 'MCP3008'

    try:
        buffers = [AnalogSensorBuffer(sensor_probe(channel=s.channel), s.bias,
                                      n_samples_buffer_rms if s.is_rms else n_samples_buffer_normal,
                                      rms_sensor=s.is_rms) for s in SENSORS]

        software_spi_mode = (probe_type is 'DummySensor') or buffers[0].software_spi
        log('ENERPI ANALOG SENSING WITH {} - (channel={}, raw_value={:.5f}); '
            'Active:{}, Software SPI:{}, #buffer_rms:{}, #buffer:{}'
            .format(probe_type, SENSORS[0].channel, buffers[0].read(), buffers[0].is_active, software_spi_mode,
                    n_samples_buffer_rms, n_samples_buffer_normal), 'debug', verbose)
        if software_spi_mode and (probe_type is 'MCP3008'):
            log('SOFTWARE_SPI --> No hardware/driver present, so bye bye...', 'error', verbose)
            raise KeyboardInterrupt
        counter_buffer_rms = counter_buffer_normal = 0

        if delta_secs_raw_capture is not None:
            max_counter_frames = delta_secs_raw_capture * n_samples_buffer
            counter_frames = 0
            buffer_values = np.zeros((n_samples_buffer, len(buffers)))
            buffer_dates = np.array([np.nan] * n_samples_buffer, dtype=dt.datetime)
            tic = time()
            while counter_frames < max_counter_frames:
                buffer_values[counter_buffer_normal, :] = [b.read() for b in buffers]
                ts = dt.datetime.now()
                buffer_dates[counter_buffer_normal] = ts
                counter_buffer_normal += 1
                counter_frames += 1
                if counter_buffer_normal == n_samples_buffer:
                    yield (buffer_dates, buffer_values)
                    counter_buffer_normal = 0
                if con_pausa:
                    sleep(max(.00001, (min_ts_ms - .05) / 1000 - (time() - tic)))
                    tic = time()
        else:
            cumsum_sensors_rms = np.zeros(sum([b.is_rms for b in buffers]), dtype=float)
            cumsum_sensors_normal = other_values = np.zeros(sum([not b.is_rms for b in buffers]), dtype=float)
            assert (cumsum_sensors_rms.shape[0] + cumsum_sensors_normal.shape[0] == len(buffers))
            stop = tic = time()
            while True:
                counter_buffer_rms += 1
                process_all_sensors = counter_buffer_rms % measure_ldr_divisor == 0
                if process_all_sensors:
                    counter_buffer_normal += 1
                    # Read instant values:
                    [b.append_reading() for b in buffers]
                    cumsum_sensors_rms += [b.mean() for b in buffers if b.is_rms]
                    cumsum_sensors_normal += [b.mean() for b in buffers if not b.is_rms]
                else:
                    # Read instant values:
                    [b.append_reading() for b in buffers if b.is_rms]
                    cumsum_sensors_rms += [b.mean() for b in buffers if b.is_rms]

                # yield & reset every delta_sampling_calc:
                ts = time()
                if (ts - stop > delta_sampling_calc - PREC_SAMPLING) and counter_buffer_rms > MIN_NUM_SAMPLES_RMS:
                    stop = ts
                    power_rms_values = np.sqrt(cumsum_sensors_rms / counter_buffer_rms) * SENSORS.rms_multiplier
                    if counter_buffer_normal > 0:
                        other_values = cumsum_sensors_normal / counter_buffer_normal
                    yield (dt.datetime.now(), *power_rms_values, *other_values,
                           counter_buffer_rms, counter_buffer_normal)
                    cumsum_sensors_rms[:] = 0
                    cumsum_sensors_normal[:] = 0
                    counter_buffer_rms = counter_buffer_normal = 0
                elif con_pausa:
                    t_sleep = (min_ts_ms - .05) / 1000 - (ts - tic)
                    if t_sleep > .00005:
                        sleep(t_sleep)
                    tic = time()
    except KeyboardInterrupt:
        log('KeyboardInterrupt en PISAMPLER: Exiting...', 'warn', verbose)
    except OSError as e:
        log('OSError en PISAMPLER: "{}". Terminando el generador con KeyboardInterrupt.'.format(e), 'error', verbose)
    except (RuntimeError, AttributeError) as e:
        log('{} en PISAMPLER: "{}". Terminando el generador.'.format(e.__class__, e), 'error', verbose)
    # Exiting
    [b.close_probe() for b in buffers]
    raise KeyboardInterrupt


def enerpi_sampler_rms(n_samples_buffer=SENSORS.n_samples_buffer_rms,
                       delta_sampling=SENSORS.delta_sec_data, min_ts_ms=SENSORS.ts_data_ms,
                       use_dummy_sensors=False, verbose=False):
    """
    Generador de valores RMS de las conexiones analógicas vía MCP3008.
        - Esta función realiza el sampling de alta frecuencia y va calculando los valores RMS con un buffer (como una
        rolling mean de ventana=n_samples_buffer). Cada "delta_sampling" s, emite un mensaje con los valores calculados.
        - Mide la potencia eléctrica aplicando la conversión de voltajes correspondiente
        - Mide la intensidad luminosa de un sensor LDR (de 0. a 1.)
        - Mide el ruido de la señal (midiendo directamente un AnalogIn desconectado)
        - Mide el valor de referencia (midiendo directamente un AnalogIn puenteado a V_ref = 3.3 V)
    Devuelve, cada ∆T, una tupla con la marca temporal y las medidas en RMS

    :param n_samples_buffer: Nº de samples tenidos en cuenta para calcular el RMS instantáneo.
    :param delta_sampling: ∆T en segundos entre envíos de información (yielding)
    :param min_ts_ms: ∆T en ms mínimo entre samples. Por defecto a 0: el máximo nº de frames que pueda computarse.
    :param use_dummy_sensors: Utiliza un generador de datos por software en lugar de usar MCP3008 (para DEMO / tests).
    :param verbose: Salida de msgs de error por sys.stdout.

    :yield: (ts_datetime, power_rms, noise_rms, counter_buffer, ldr_rms)
    """
    return _sampler(n_samples_buffer=n_samples_buffer,
                    delta_sampling=delta_sampling, min_ts_ms=min_ts_ms,
                    use_dummy_sensors=use_dummy_sensors, verbose=verbose)


def enerpi_raw_sampler(delta_secs=20, n_samples_buffer=SENSORS.n_samples_buffer_rms, min_ts_ms=0,
                       use_dummy_sensors=False, verbose=True):
    """
    Generador de valores en bruto de las conexiones analógicas vía MCP3008 (sampling de alta frecuencia)

    :param delta_secs: Nº de samples tenidos en cuenta para calcular el RMS instantáneo.
    :param n_samples_buffer: Nº de samples tenidos en cuenta para hacer el yielding.
    :param min_ts_ms: ∆T en ms mínimo entre samples. Por defecto a 0: el máximo nº de frames que pueda computarse.
    :param use_dummy_sensors: Utiliza un generador de datos por software en lugar de usar MCP3008 (para DEMO / tests).
    :param verbose: Salida de msgs de error por sys.stdout.

    :yield: (ts_datetime, power_rms, noise_rms, counter_buffer, ldr_rms)
    """
    return _sampler(delta_secs_raw_capture=delta_secs,
                    n_samples_buffer=n_samples_buffer, min_ts_ms=min_ts_ms,
                    use_dummy_sensors=use_dummy_sensors, reset_bias=True, verbose=verbose)



# -*- coding: utf-8 -*-
from flask import request, redirect, url_for, render_template, jsonify, abort
import json
from threading import Timer
import os
import sys
from enerpi.base import log
from enerpi.api import enerpi_data_catalog
from enerpiplot.plotbokeh import get_bokeh_version
from enerpiweb import app, auto, WITH_ML_SUBSYSTEM
from enerpiweb.rt_stream import stream_is_alive
from enerpiweb.forms import DummyForm


BOKEH_VERSION = get_bokeh_version()

if WITH_ML_SUBSYSTEM:
    # from enerpiweb.views_labeling import *
    # TODO Integración con las vistas de enerpiprocess (separadas en otro project por dependencias, por ahora)
    pass
# else:
#     @app.route('/learning')
#     def index_learning():
#         return redirect(url_for('control'))


#############################
# INDEX & ROUTES
#############################
@app.route('/control', methods=['GET'])
@auto.doc()
def control():
    """
    Admin Control Panel with links to LOG viewing/downloading, hdf_stores download, ENERPI config editor, etc.

    """
    is_sender_active, last = stream_is_alive()
    after_sysop = request.args.get('after_sysop', '')
    alerta = request.args.get('alerta', '')
    if alerta:
        alerta = json.loads(alerta)
    cat = enerpi_data_catalog(check_integrity=False)
    df = cat.tree
    if df is not None:
        df = df[df.is_cat & df.is_raw].sort_values(by='ts_ini', ascending=False)
        paths_rel = [(os.path.basename(p), t0.strftime('%d/%m/%y'), tf.strftime('%d/%m/%y'), n)
                     for p, t0, tf, n in zip(df['st'], df['ts_ini'], df['ts_fin'], df['n_rows'])]
    else:
        paths_rel = []
    form_operate = DummyForm()
    return render_template('control_panel.html',
                           d_catalog={'path_raw_store': os.path.join(cat.base_path, cat.raw_store),
                                      'path_catalog': os.path.join(cat.base_path, cat.catalog_file),
                                      'ts_init': cat.min_ts, 'ts_catalog': cat.index_ts},
                           d_last_msg=last, is_sender_active=is_sender_active, list_stores=paths_rel,
                           form_operate=form_operate, after_sysop=after_sysop,
                           alerta=alerta)


@app.route('/api/help', methods=['GET'])
def api_help():
    """
    Documentation page generated with 'Autodoc', with custom template;
    or json response with server routes (with ?json=true)

    """
    w_json = request.args.get('json', False)
    if w_json:
        endpoints = [rule.rule for rule in app.url_map.iter_rules()
                     if rule.endpoint != 'static']
        return jsonify(dict(api_endpoints=endpoints))
    return auto.html(template='doc/api_help.html', title='enerPI Help')


@app.route('/api/bokehplot', methods=['GET'])
@auto.doc()
def bokehplot():
    """
    Base webpage for query & show bokeh plots of ENERPI data

    """
    return render_template('bokeh_plot.html', url_stream_bokeh=url_for('bokeh_buffer'), b_version=BOKEH_VERSION)


@app.route('/', methods=['GET'])
def base_index():
    """
    Redirects to 'index', with real-time monitoring tiles of ENERPI sensors

    """
    return redirect(url_for('index'))


#############################
# ENERPI SERVER COMMAND
#############################
@app.route('/api/restart/<service>', methods=['POST'])
@auto.doc()
def startstop(service='enerpi_start'):
    """
    Endpoint for control ENERPI in RPI. Only for dev mode.
    It can restart the ENERPI daemon logger or even reboot the machine for a fresh start after a config change.

    :param service: service id ('enerpi_start/stop' for operate with the logger, or 'machine' for a reboot)

    """
    def _system_operation(command):
        log('SYSTEM_OPERATION CMD: "{}"'.format(command), 'debug', False)
        os.system(command)

    form = DummyForm()
    cmd = msg = alert = None
    if form.validate_on_submit():
        if service == 'enerpi_start':
            python_pathbin = os.path.dirname(sys.executable)
            cmd = '{}/enerpi-daemon start'.format(python_pathbin)
            msg = 'Starting ENERPI logger from webserver... ({})'.format(cmd)
            alert = 'warning'
        elif service == 'enerpi_stop':
            python_pathbin = os.path.dirname(sys.executable)
            cmd = '{}/enerpi-daemon stop'.format(python_pathbin)
            msg = 'Stopping ENERPI logger from webserver... ({})'.format(cmd)
            alert = 'danger'
        elif service == 'machine':
            cmd = 'reboot now'
            msg = 'Rebooting! MACHINE... see you soon... ({})'.format(cmd)
            alert = 'danger'
        if cmd is not None:
            log(msg, 'debug', False)
            t = Timer(.5, _system_operation, args=(cmd,))
            t.start()
            return redirect(url_for('control', after_sysop=True,
                                    alerta=json.dumps({'alert_type': alert, 'texto_alerta': msg})))
    return abort(500)



# -*- coding: utf-8 -*-
import logging
import numpy as np
import os
import pandas as pd
from subprocess import check_output
import sys
import threading
from time import sleep

from enerpi.base import DATA_PATH, log
from enerpi.enerpimeter import enerpi_logger, enerpi_raw_data
from enerpi.prettyprinting import *

"""
* ENERPI Calibration utils *

Use these for:
    1. Checking electrical-net frequency (50Hz in Europe)
    2. Checking real value of voltage divisors in RMS sensing circuit schematics (without the SCT probes)
    3. Test and compare CPU usage vs sampling frequency, in order to select the optimum value for your machine.

    * Summary results for MBP:

                 CPU_mean  CPU_min  CPU_max  CPU_median  MEM_mean  MEM_min  MEM_max
    ts_data_ms
    0.00        77.420000      0.8    102.6       99.60       0.4      0.4      0.4
    0.50        32.675000      0.0     47.0       40.80       0.4      0.4      0.4
    0.75        23.096667      0.0     35.7       28.60       0.4      0.4      0.4
    1.00        13.998333      0.9     24.6       15.95       0.4      0.4      0.4
    2.00        10.510000      0.9     15.8       13.70       0.4      0.4      0.4
    3.00         7.986667      0.0     11.8       10.10       0.4      0.4      0.4
    4.00         6.986667      1.0     10.3        8.55       0.4      0.4      0.4
    5.00         5.805000      0.0      9.2        7.00       0.4      0.4      0.4
    6.00         4.811667      1.0      7.0        5.70       0.4      0.4      0.4
    7.00         4.653333      1.0      7.0        5.50       0.4      0.4      0.4
    8.00         4.345000      0.0      7.0        5.15       0.4      0.4      0.4
    9.00         3.731667      0.9      6.2        4.40       0.4      0.4      0.4
    10.00        3.478333      0.8      5.3        3.90       0.4      0.4      0.4
    11.00        3.350000      0.9      5.2        3.80       0.4      0.4      0.4
    12.00        3.193333      0.9      5.3        3.60       0.4      0.4      0.4

                 CPU_mean  CPU_min  CPU_max  CPU_median  MEM_mean  MEM_min  MEM_max
    ts_data_ms
    0.00        21.000000      0.8    105.7        1.20   0.48125  0.48125  0.48125
    0.50        26.825000      2.5     40.6       28.50   0.50000  0.50000  0.50000
    0.75        19.778125      3.4     27.4       20.60   0.50000  0.50000  0.50000
    1.00        17.768750      3.5     23.9       18.50   0.50000  0.50000  0.50000
    2.00        10.556250      2.8     13.0       11.10   0.50000  0.50000  0.50000
    3.00         7.671875      1.6     15.3        8.10   0.50000  0.50000  0.50000
    4.00         6.050000      1.6      8.6        6.15   0.50000  0.50000  0.50000
    5.00         5.403125      0.1     14.7        5.40   0.50000  0.50000  0.50000
    6.00         4.971875      0.0      6.5        5.25   0.50000  0.50000  0.50000
    7.00         4.759375      1.9      7.7        4.95   0.50000  0.50000  0.50000
    8.00         4.834375      1.9     11.9        4.60   0.50000  0.50000  0.50000
    9.00         4.071875      1.2      7.0        4.10   0.50000  0.50000  0.50000
    10.00        4.025000      1.6      7.9        4.00   0.50000  0.50000  0.50000
    11.00        3.190625      0.9      5.0        3.25   0.50000  0.50000  0.50000
    12.00        3.534375      1.4      9.5        3.40   0.50000  0.50000  0.50000
    * Summary results for RPI2:

                 CPU_mean  CPU_min  CPU_max  CPU_median  MEM_mean   MEM_min   MEM_max
    ts_data_ms
    0.00        94.865000     89.7    108.0       93.75  6.091667  6.091667  6.091667
    0.50        90.376667     87.4     93.1       90.20  7.693333  7.693333  7.693333
    0.75        87.806667     86.1     89.1       87.80  7.980000  7.980000  7.980000
    1.00        86.298333     84.7     87.3       86.30  8.725000  8.725000  8.725000
    2.00        83.155000     79.7     85.8       83.15  9.428333  9.428333  9.428333
    3.00        78.111667     74.8     80.7       78.15  9.781667  9.781667  9.781667
    4.00        73.433333     70.7     75.7       73.60  8.125000  8.125000  8.125000
    5.00        69.323333     66.9     71.4       69.40  7.895000  7.895000  7.895000
    6.00        65.856667     63.7     67.6       66.00  9.166667  9.166667  9.166667
    7.00        62.921667     61.0     64.6       63.00  9.876667  9.876667  9.876667
    8.00        60.415000     58.7     61.9       60.50  8.333333  8.333333  8.333333
    9.00        58.241667     56.7     59.5       58.40  8.673333  8.673333  8.673333
    10.00       56.336667     54.9     57.4       56.40  9.263333  9.263333  9.263333
    11.00       54.678333     53.5     55.7       54.80  9.963333  9.963333  9.963333
    12.00       53.156667     52.1     54.1       53.20  8.618333  8.618333  8.618333

                 CPU_mean  CPU_min  CPU_max  CPU_median  MEM_mean  MEM_min  MEM_max
    ts_data_ms
    0.00        52.296875     52.1     52.5       52.30       8.6      8.6      8.6
    0.50        53.200000     53.1     53.3       53.20       7.8      7.8      7.8
    0.75        53.534375     53.4     53.7       53.50       7.8      7.8      7.8
    1.00        53.531250     53.4     53.6       53.50       7.8      7.8      7.8
    2.00        53.053125     52.8     53.4       53.00       7.8      7.8      7.8
    3.00        52.290625     51.9     52.7       52.30       7.8      7.8      7.8
    4.00        51.456250     51.1     51.9       51.45       7.8      7.8      7.8
    5.00        50.568750     50.1     51.0       50.55       7.8      7.8      7.8
    6.00        49.706250     49.3     50.1       49.70       7.8      7.8      7.8
    7.00        48.859375     48.5     49.3       48.90       7.8      7.8      7.8
    8.00        48.025000     47.6     48.4       48.00       7.8      7.8      7.8
    9.00        47.231250     46.9     47.6       47.20       7.8      7.8      7.8
    10.00       46.443750     46.1     46.9       46.45       7.8      7.8      7.8
    11.00       45.690625     45.3     46.0       45.70       7.8      7.8      7.8
    12.00       44.953125     44.6     45.3       44.90       7.8      7.8      7.8


"""

PATH_ST_CALIB = os.path.join(DATA_PATH, 'raw_data_calibration.h5')
LOG_FILE_CALIB = os.path.join(DATA_PATH, 'calibration.log')


def _measure_cpu_psaux(result_data, key, ts_data_ms, interval=5, freq_s=.25):
    pid = os.getpid()
    cmd = ['ps', 'aux']
    out = check_output(cmd).decode().splitlines()
    columns = out[0].split()
    ncols = len(columns)
    idx_pid_col = columns.index('PID')
    result = []
    for i in range(interval * int(1 / freq_s)):
        out = check_output(cmd).decode().splitlines()
        result += list(map(lambda x: x.split()[:ncols - 1] + [' '.join(x.split()[ncols - 1:])],
                           filter(lambda l: l.split()[idx_pid_col] == str(pid), out[1:])))
        sleep(freq_s)
    df_ps = pd.DataFrame(result, columns=columns)
    df_ps['ts_data_ms'] = ts_data_ms
    result_data[key] = df_ps


def ascii_histogram(raw, bins=30, total_width=80):
    """
    Plot histogram in ASCII mode

    :param raw: pd.Series or pd.DataFrame with timeseries index
    :param bins: # of bins for the histogram (= nº of plot bar lines)
    :param total_width: console width
    :return: np.array with ∆T in ms

    """
    def _make_bargraph(value, label, v_max=4000, ancho_disp_bar=60):
        n_resto = 5
        v_bar = value * ancho_disp_bar / v_max
        n_big = int(v_bar)
        n_little = int(round((v_bar - int(v_bar)) / (1 / n_resto)))
        if n_little == n_resto:
            n_little = 0
            n_big += 1
        line = label + '◼︎' * n_big + '⇡︎' * n_little
        return line

    delta_ms = np.array((raw.index[1:].values - raw.index[:-1].values).astype(int) / 1e6)
    print_magenta('Histogram of ∆T sampling: min={:.3f}; max={:.3f}; mean={:.3f}; median={:.3f}'
                  .format(delta_ms.min(), delta_ms.max(), delta_ms.mean(), np.median(np.round(delta_ms, 3))))
    n_samples, ranges = np.histogram(delta_ms, bins=bins)
    max_n_samples = max(n_samples)
    mask_label = '{:.2f}->{:.2f}ms => {:5d}: '
    len_label = 10 + 2 * 3 + 5
    assert total_width - len_label > 10
    lines = [_make_bargraph(v, mask_label.format(a, b, v), v_max=max_n_samples, ancho_disp_bar=total_width - len_label)
             for v, a, b in zip(n_samples, ranges[:-1], ranges[1:])]
    for l in lines:
        print_red(l)
    return delta_ms


def _cpu_use_raw_sampling_ts(results, path_st, ts_data_ms, delta_secs=10,
                             show_histograms=False, use_dummy_sensors=True, **kwargs_logger):
    # func_enerpi_logger(results, path_st, ts_data_ms, delta_secs, show_histograms, **kwargs_logger)
    key = 'raw_ts_{}'.format(str(ts_data_ms).replace('.', '_'))
    print_secc('CPU usage (raw sampling with ∆ts = {} ms, for {} seconds)'.format(ts_data_ms, delta_secs))
    timer = threading.Thread(target=_measure_cpu_psaux, args=(results, key, ts_data_ms, delta_secs))
    timer.start()
    raw = enerpi_raw_data(path_st, key_save=key, sampling_ms=ts_data_ms, delta_secs=delta_secs,
                          use_dummy_sensors=use_dummy_sensors, **kwargs_logger)
    timer.join()
    # Raw sampling stats plots:
    if show_histograms:
        _ = ascii_histogram(raw, bins=20, total_width=120)


# noinspection PyUnusedLocal
def _cpu_use_normal_sampling_ts(results, path_st, ts_data_ms, delta_secs=10,
                                show_histograms=False, use_dummy_sensors=True, **kwargs_logger):
    # func_enerpi_logger(results, path_st, ts_data_ms, delta_secs, show_histograms, **kwargs_logger)
    key = 'rms_ts_{}'.format(str(ts_data_ms).replace('.', '_'))
    print_secc('CPU usage (raw sampling with ∆ts = {} ms, for {} seconds)'.format(ts_data_ms, delta_secs))
    timer = threading.Thread(target=_measure_cpu_psaux, args=(results, key, ts_data_ms, delta_secs))
    timer.start()
    enerpi_logger(path_st, is_demo=use_dummy_sensors, sampling_ms=ts_data_ms, timeout=delta_secs, **kwargs_logger)
    timer.join()


def _cpu_use_sampling_ts(func_enerpi_logger,
                         ts_ms_values=(0, 0.25, 0.5, 1, 2, 3, 4, 5, 6, 7, 8),
                         delta_secs=10, show_histograms=True, use_dummy_sensors=True, **kwargs_logger):
    """
    Test CPU usage in raw sampling with multiple ts_data values, and summarize results

    :param ts_ms_values:
    :param delta_secs:
    :param use_dummy_sensors:
    :return: summary of test results

    """
    # CPU usage:
    results = {}
    path_st = PATH_ST_CALIB
    for ts_data_ms in ts_ms_values:
        func_enerpi_logger(results, path_st, ts_data_ms, delta_secs,
                           show_histograms, use_dummy_sensors, **kwargs_logger)
    results_ps_raw = pd.DataFrame(pd.concat(results.values())).sort_values(by='TIME')
    results_ps_raw['%CPU'] = results_ps_raw['%CPU'].apply(lambda x: float(x.replace(',', '.')))
    results_ps_raw['%MEM'] = results_ps_raw['%MEM'].apply(lambda x: float(x.replace(',', '.')))
    print_info(results_ps_raw)
    gb = results_ps_raw.groupby('ts_data_ms')
    summary = pd.concat([gb['%CPU'].mean().rename('CPU_mean'),
                         gb['%CPU'].min().rename('CPU_min'),
                         gb['%CPU'].max().rename('CPU_max'),
                         gb['%CPU'].median().rename('CPU_median'),
                         gb['%MEM'].mean().rename('MEM_mean'),
                         gb['%MEM'].mean().rename('MEM_min'),
                         gb['%MEM'].mean().rename('MEM_max')],
                        axis=1)
    print_ok(summary)
    return pd.DataFrame(summary)


def cpu_use_raw_sampling_ts(ts_ms_values=(0, 0.25, 0.5, 1, 2, 3, 4, 5, 6, 7, 8),
                            delta_secs=10, use_dummy_sensors=True,
                            show_histograms=True, verbose=False):
    """
    Test CPU usage in raw sampling with multiple ts_data values, and summarize results

    :param ts_ms_values:
    :param delta_secs:
    :param use_dummy_sensors:
    :param show_histograms:
    :param verbose:
    :return: summary of test results

    """
    return _cpu_use_sampling_ts(_cpu_use_raw_sampling_ts,
                                ts_ms_values=ts_ms_values, delta_secs=delta_secs, use_dummy_sensors=use_dummy_sensors,
                                show_histograms=show_histograms, roll_time=1, verbose=verbose)


def cpu_use_normal_sampling_ts(ts_ms_values=(0, 0.25, 0.5, 1, 2, 3, 4, 5, 6, 7, 8),
                               delta_secs=10, use_dummy_sensors=True):
    """
    Test CPU usage in raw sampling with multiple ts_data values, and summarize results

    :param ts_ms_values:
    :param delta_secs:
    :param use_dummy_sensors:
    :return: summary of test results

    """
    return _cpu_use_sampling_ts(_cpu_use_normal_sampling_ts,
                                ts_ms_values=ts_ms_values, delta_secs=delta_secs, use_dummy_sensors=use_dummy_sensors,
                                roll_time=1, delta_sampling=1, verbose=True)


if __name__ == '__main__':
    pd.set_option('display.width', 160)
    logging.basicConfig(filename=LOG_FILE_CALIB, level='DEBUG', datefmt='%d/%m/%Y %H:%M:%S',
                        format='%(levelname)s [%(filename)s_%(funcName)s] - %(asctime)s: %(message)s')
    is_mac = sys.platform == 'darwin'
    log('ENERPI CALIBRATION UTIL ON {}'.format('macOS' if is_mac else 'RPI'), 'magenta', True)

    # 3. Test and compare CPU usage vs sampling frequency.
    ts_test = (0, 0.5, 0.75, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12)

    res_ts = cpu_use_normal_sampling_ts(ts_ms_values=ts_test, delta_secs=15, use_dummy_sensors=is_mac)
    res_ts.to_hdf(PATH_ST_CALIB, 'summary_cpu_normal')

    res_raw_ts = cpu_use_raw_sampling_ts(ts_ms_values=ts_test, delta_secs=8, use_dummy_sensors=is_mac,
                                         show_histograms=True, verbose=False)
    res_raw_ts.to_hdf(PATH_ST_CALIB, 'summary_cpu_raw')



# -*- coding: utf-8 -*-
import pandas as pd
from enerpi.hdftscat import HDFTimeSeriesCatalog
from enerpi.base import log


'''
    @staticmethod
    def process_data(data):
        # Implementar
        return data

    @staticmethod
    def process_data_summary(data):
        # Implementar
        return data, data

    @staticmethod
    def process_data_summary_extra(data):
        # Implementar
        return data, data, None
'''

DELTA_MAX_SUMMARY_BFILL = 120  # pd.Timedelta('2min')
# TODO Terminar Doc


def _compress_data(data, sensors):
    """
    Down-casting of raw data to minimize disk usage.

    RMS sensor data is reduced to float32
    MEAN sensor data take integer values from 0 to 1000 (int16)
    Ref counters of rms & mean sensors also take integer values (int16)
    :param data: :pd.DataFrame: raw data
    :return: :pd.DataFrame: compressed raw data
    """
    if (data is not None) and not data.empty:
        c1, c2, c3 = sensors.columns_sensors_rms, sensors.columns_sensors_mean, [sensors.ref_rms, sensors.ref_mean]
        try:
            data_c = data[c1 + c2 + c3].copy()
        except KeyError as e:
            c1, c2, c3 = [list(filter(lambda x: x in data, c)) for c in [c1, c2, c3]]
            log('KeyError in _compress_data: {}; existent columns: {}'.format(e, [c1, c2, c3]), 'error', False)
            data_c = data[c1 + c2 + c3].copy()
        for c in c1:
            data_c[c] = data_c[c].astype('float32')
        for c in c2:
            if data_c[c].dtype != 'int16':
                data_c[c] = pd.Series(1000. * data_c[c]).round().astype('int16')
        for c in c3:
            data_c[c] = data_c[c].astype('int16')
        return data_c
    return data


def _process_data(data, sensors):
    """
    Calculate summary data (consumption and some stats of enerpi data...)

    :param data: :pd.DataFrame: archived data
    :return:
    """
    if data is not None and not data.empty:
        data = data.copy()
        col_delta = 'delta'
        col_cons = 'kWh'
        data[col_delta] = pd.Series(data.index).diff().fillna(method='bfill').dt.total_seconds().values
        data.loc[data[col_delta] > DELTA_MAX_SUMMARY_BFILL, col_delta] = DELTA_MAX_SUMMARY_BFILL
        cols_summary = []
        for i, c in enumerate(sensors.columns_sensors_rms):
            if i == 0:
                col_s = col_cons
            else:
                col_s = '{}_{}'.format(col_cons, c)
            cols_summary.append(col_s)
            data[col_s] = data[c] * data.delta / 3600
        cols_rs = list(sensors.columns_sensors_rms) + cols_summary + [col_delta]
        resampler = data[cols_rs].resample('1h', label='left')
        consumo = resampler[cols_summary].sum().fillna(0.).astype('float32')
        consumo /= 1000.
        consumo['t_ref'] = pd.Series(resampler[col_delta].sum() / 3600).astype('float32')
        consumo['p_max'] = resampler[sensors.main_column].max().round(0).astype('float16')
        consumo['p_mean'] = resampler[sensors.main_column].mean().round(0).astype('float16')
        consumo['p_min'] = resampler[sensors.main_column].min().round(0).astype('float16')
        data.drop([col_delta, col_cons], axis=1, inplace=True)
        return consumo
    return None


class EnerpiCatalog(HDFTimeSeriesCatalog):

    def __init__(self, sensors=None, **kwargs):
        if sensors is None:
            from enerpi.base import reload_config
            reload_config()
            from enerpi.base import SENSORS
            sensors = SENSORS
        self.sensors = sensors
        HDFTimeSeriesCatalog.__init__(self, **kwargs)

    @staticmethod
    def is_raw_data(data):
        for dt in data.dtypes:
            if dt == 'float64':
                return True
        return False

    def process_data(self, data):
        return _compress_data(data, self.sensors)

    def process_data_summary(self, data):
        return _process_data(data, self.sensors)

    def process_data_summary_extra(self, data):
        consumo = _process_data(data, self.sensors)
        return consumo, None



# -*- coding: utf-8 -*-
import datetime as dt
import os
import pandas as pd
import requests
import tempfile
# noinspection PyUnresolvedReferences
from enerpi.base import CONFIG, SENSORS, DATA_PATH, FILE_LOGGING, LOGGING_LEVEL, log, timeit
# noinspection PyUnresolvedReferences
from enerpi.database import init_catalog, get_ts_last_save, extract_log_file, delete_log_file, HDF_STORE
# noinspection PyUnresolvedReferences
from enerpi.enerpimeter import receiver_msg_generator, enerpi_raw_data, msg_to_dict
# noinspection PyUnresolvedReferences
from enerpi.iobroadcast import get_encryption_key, get_codec
from enerpi.hdftscat import get_catalog_paths


def enerpi_receiver_generator(verbose=False, n_msgs=None):
    """
    Generator of broadcasted values by ENERPI Logger.

    It can be used by any machine in the same network as the ENERPI Logger. It decrypts the encrypted broadcast and
    returns a dict of vars - values.
    Used by the webserver for read & stream real-time values.

    :param verbose: :bool: Log to stdout
    :param n_msgs: :int: # of msgs to receive (unlimited by default).
    :return: :dict:
    """
    gen = receiver_msg_generator(verbose=verbose, n_msgs=n_msgs)
    count = 0
    while True:
        try:
            msg, _t1, _t2 = next(gen)
            yield msg_to_dict(msg)
            count += 1
        except StopIteration:
            log('EXIT from enerpi_receiver_generator. StopIteration in msg #{}'.format(count), 'error', verbose)
            break
    return None


def enerpi_default_config():
    """
    Default configuration for ENERPI Data Catalog, read from INI file.

    :return: :dict: parameters
    """
    conf = {'store': HDF_STORE,
            'DATA_PATH': DATA_PATH,
            'delta': SENSORS.delta_sec_data,
            'window': SENSORS.rms_roll_window_sec,
            'ts': SENSORS.ts_data_ms,
            'LOGGING_LEVEL': LOGGING_LEVEL,
            'FILE_LOGGING': FILE_LOGGING}
    return conf


def enerpi_data_catalog(check_integrity=False, **kwargs):
    """
    Get ENERPI data catalog for access & operation.

    :param check_integrity: :bool: False by default. If true, checks integrity and generates / updates data index.
    :param kwargs: :dict:
    :return: :EnerpiCatalog:
    """
    return init_catalog(check_integrity=check_integrity, **kwargs)


@timeit('remote_data_get', verbose=True)
def remote_data_get(t0, tf=None,
                    enerpi_ip='192.168.1.52', port=80,
                    prefix_remote_enerpi='/enerpi', verbose=True):
    """
    Query a ENERPI catalog in a remote machine with enerpiweb running.

    :param t0: start of slice
    :param tf:  end of slice (or None for end = now)
    :param enerpi_ip: IP of the remote machine
    :param port: PORT of the remote enerpiweb server
    :param prefix_remote_enerpi: URL prefix of the remote enerpiweb server
    :param verbose: :bool: verbose mode
    :return: dict of (data_key, pd.DataFrame of sliced values) pairs

    """
    def _request_store(url_api_download_st, dest_path=None):
        r = requests.get(url_api_download_st)
        if r.ok and (r.headers['Content-Type'] == 'application/octet-stream'):
            if verbose:
                size_kb = int(r.headers['Content-Length']) / 1024
                date_mod = pd.Timestamp(r.headers['Last-Modified']).tz_convert('Europe/Madrid')
                log('STORE Downloaded from "{}" --> {:.2f} KB, mod={:%d/%m/%y %H:%M:%S}'
                    .format(url_api_download_st, size_kb, date_mod), 'ok', True, False)
            # delete_filest = False
            if dest_path is None:
                dest_path = os.path.join(tempfile.gettempdir(), 'temp_store.h5')
                # delete_filest = True
            with open(dest_path, 'wb') as f:
                f.write(r.content)
            if verbose:
                local_size = os.path.getsize(dest_path) / 1024
                local_date_mod = dt.datetime.fromtimestamp(os.path.getmtime(dest_path))
                log('STORE NOW IN LOCAL DISK AT "{}", {:.2f} KB, mod={:%d/%m/%y %H:%M:%S}'
                    .format(dest_path, local_size, local_date_mod), 'magenta', True, False)
            # if delete_filest:
            # Return content & remove hdf temporal file store:
            with pd.HDFStore(dest_path, 'r') as st:
                data = {k: st[k] for k in st.keys()}
                if verbose:
                    log('HDF Store downloaded:\n{}'.format(st), 'ok', True, False)
            os.remove(dest_path)
            return data

    paths = get_catalog_paths(t0, tf)
    url_mask = 'http://{}:{}{}/api/hdfstores/'.format(enerpi_ip, port, prefix_remote_enerpi) + '{}'
    data_stores = []
    for p in paths:
        url = url_mask.format(os.path.split(p)[-1])
        if verbose:
            log('REQUEST HDF STORE AT: {}'.format(url), 'info', True, False)
        data_stores.append(_request_store(url, dest_path=None))
    keys = set([k for d in data_stores for k in d.keys()])
    data_out = {k: pd.DataFrame(pd.concat([data[k].loc[t0:tf] for data in data_stores
                                           if k in data.keys()])).sort_index()
                for k in keys}
    return data_out



# -*- coding: utf-8 -*-
import locale
import numpy as np
import pandas as pd
import bokeh
from bokeh.plotting import figure
from bokeh.embed import components
from bokeh.models import HoverTool, ColumnDataSource, LinearAxis, Range1d
from bokeh.io import reset_output
from enerpi.base import SENSORS, CUSTOM_LOCALE


ROUND_W = 500
ROUND_KWH = .5

COLS_SENSORS_RMS = SENSORS.columns_sensors_rms
COLS_SENSORS_MEAN = SENSORS.columns_sensors_mean
COL_MAIN_RMS = COLS_SENSORS_RMS[0]
COL_MAIN_MEAN = COLS_SENSORS_MEAN[0]

COL_TS = SENSORS.ts_column
COLS_DATA = COLS_SENSORS_RMS + COLS_SENSORS_MEAN + [SENSORS.ref_rms]
B_SENSORS_RMS = [True] * len(COLS_SENSORS_RMS) + [False] * len(COLS_SENSORS_MEAN) + [True]
# assert len(B_SENSORS_RMS) == len(COLS_DATA) - 1
COLORS_DATA = [SENSORS[s].color for s in COLS_DATA[:-1]] + ['#FF0077']
UNITS_DATA = [SENSORS[s].unit for s in COLS_DATA[:-1]] + ['']
FMT_TOOLTIP_DATA = ['{0}'] * len(COLS_SENSORS_RMS) + ['{0.0}'] * len(COLS_SENSORS_MEAN) + ['{0}']
LABELS_DATA = SENSORS.descriptions(COLS_DATA)

# TODO Adaptar summary plots
COLS_DATA_KWH = ['kWh', 'p_max', 'p_min', 't_ref']
COLORS_DATA_KWH = ['#8C27D3', '#972625', '#f4af38', '#8C27D3']
UNITS_DATA_KWH = ['kWh', 'W', 'W', '']
LABELS_DATA_KWH = ['Consumption', 'Max Power', 'Min Power', 'Sampled']
FMT_TOOLTIP_DATA_KWH = ['{0.000}', '{0}', '{0}', '{0.000}']

locale.setlocale(locale.LC_ALL, CUSTOM_LOCALE)
# locale.setlocale(locale.LC_ALL, locale.getlocale())
# locale.setlocale(locale.LC_ALL, CONFIG.get('ENERPI_SAMPLER', 'LOCALE', fallback='{}.{}'.format(*locale.getlocale())))

TOOLS = "pan,xwheel_zoom,ywheel_zoom,box_zoom,reset,save,crosshair"
P_WIDTH = 900
P_HEIGHT = 500

HTML_TROW = """<tr><td style="font-size: 15px; font-weight: bold;">
<span class="bk-tooltip-color-block" style="background-color: {2}"></span>{0}</td>
<td style="font-size: 16px; font-weight: bold; color: {2};">{1}</td></tr>"""
TOOLTIP_ROWS = """<tr><td style="font-size: 15px; font-weight: bold;">Time:</td>
<td style="font-size: 15px; font-weight: bold;">@time</td></tr>"""


def _append_hover(df, delta_min=0, multi_day=False):
    if multi_day:
        df['time'] = [(x + pd.Timedelta(minutes=delta_min)).strftime("%-d/%b %H:%M:%S") for x in df.index]
    else:
        df['time'] = [(x + pd.Timedelta(minutes=delta_min)).strftime("%H:%M:%S.%f")[:-3] for x in df.index]
    return df


def _titulo_rango_temporal(extremos_dia):
    n_days = (extremos_dia[1] - extremos_dia[0]).days
    if n_days > 1:
        return "{:%-d de %B'%y} --> {:%-d de %B'%y}".format(extremos_dia[0], extremos_dia[1])
    else:
        return '{:%A %-d de %B de %Y}'.format(extremos_dia[0])


def _get_extremos_dia(df):
    extremos = [x.replace(microsecond=0) for x in df.index[[0, -1]]]
    # extremos_dia = [pd.Timestamp(x.date() + pd.Timedelta(days=x.hour // 12)) for x in extremos]
    # return [extremos[0].date(), (extremos[1] - pd.Timedelta(minutes=1) + pd.Timedelta(days=1)).date()]
    return extremos


def get_bokeh_version():
    """
    For jinja2 templating (css & js src's)
    """
    return bokeh.__version__


def _return_html_comps(plots):
    script, divs = components(plots)
    reset_output()
    return script, divs, get_bokeh_version()


def _get_figure_plot(extremos, y_range, **fig_kwargs):
    kwargs = dict(x_range=extremos, y_range=y_range,
                  tools=TOOLS, active_drag=None, plot_width=P_WIDTH, plot_height=P_HEIGHT,
                  x_axis_type="datetime", toolbar_location="right", toolbar_sticky=False,  # "above",
                  title=_titulo_rango_temporal(extremos), responsive=True, **fig_kwargs)
    return figure(**kwargs)


def _format_axis_plot(p, color, label, unit):
    # Axis formatting
    p.axis.major_tick_in = 7
    p.axis.minor_tick_in = 3
    p.axis.major_tick_out = p.axis.minor_tick_out = 0
    p.title.text_font_size = '12pt'
    if unit:
        p.yaxis.axis_label = '{} ({})'.format(label, unit)
    else:
        p.yaxis.axis_label = label
    p.yaxis.axis_label_text_font_size = "11pt"
    p.yaxis.axis_line_color = p.yaxis.major_label_text_color = p.yaxis.axis_label_text_color = color
    p.yaxis.major_tick_line_color = p.yaxis.minor_tick_line_color = color


def _format_legend_plot(p):
    # Legend formatting
    p.legend.location = "top_left"
    p.legend.label_text_font_size = "7pt"
    p.legend.background_fill_alpha = .5
    p.legend.label_height = 12
    p.legend.glyph_height = 10
    p.legend.label_standoff = 5
    p.legend.padding = 8
    p.legend.spacing = 3


def _plot_patch_series(fig_plot, pd_series, **kwargs_patch):
    zero = np.array([0])
    df_patch = pd.Series(pd_series.fillna(method='pad', limit=2).fillna(0).round(2))
    x = np.append(np.insert(df_patch.index.values, 0, df_patch.index.values[0]), df_patch.index.values[-1])
    y = np.append(np.insert(df_patch.values, 0, zero), zero)
    fig_plot.patch(x, y, **kwargs_patch)


def _plot_bokeh_raw(data_plot, **fig_kwargs):
    # Bokeh does not work very well!! with timezones:
    data_plot = data_plot.tz_localize(None)
    for c in COLS_SENSORS_MEAN:
        data_plot[c] /= 10.
    y_range = [0, max(500, int(np.ceil(data_plot[[COL_MAIN_RMS, SENSORS.ref_rms]].max().max() / ROUND_W) * ROUND_W))]
    minmax_ejes = [y_range, [0, 100]]

    extremos = _get_extremos_dia(data_plot)
    ejes = [COL_MAIN_RMS, COL_MAIN_MEAN]
    pos_smean = len(COLS_SENSORS_RMS)

    # Figure
    p = _get_figure_plot(extremos, minmax_ejes[0], **fig_kwargs)

    tooltip_rows = TOOLTIP_ROWS
    for c, color, unit, label, fmt in zip(COLS_DATA, COLORS_DATA, UNITS_DATA, LABELS_DATA, FMT_TOOLTIP_DATA):
        tooltip_rows += HTML_TROW.format('{}:'.format(label), '@{}{} {}'.format(c, fmt, unit), color)
    p.add_tools(HoverTool(tooltips='<div><table>{}</table></div>'.format(tooltip_rows)))

    # Axis formatting
    _format_axis_plot(p, COLORS_DATA[0], LABELS_DATA[0], UNITS_DATA[0])
    if len(ejes) > 1:
        positions = (['right', 'left'] * 2)[:len(ejes[1:])]
        for extra_eje, pos, minmax, color, label, unit in zip(ejes[1:], positions, minmax_ejes[1:],
                                                              COLORS_DATA[pos_smean:], LABELS_DATA[pos_smean:],
                                                              UNITS_DATA[pos_smean:]):
            p.extra_y_ranges[extra_eje] = Range1d(*minmax)
            axkw = dict(y_range_name=extra_eje,
                        axis_label='{} ({})'.format(label, unit),
                        axis_label_text_font_size="10pt", axis_line_color=color,
                        major_label_text_color=color, axis_label_text_color=color,
                        major_tick_line_color=color, minor_tick_line_color=color,
                        major_tick_in=5, major_tick_out=0, minor_tick_in=3, minor_tick_out=0)
            p.add_layout(LinearAxis(**axkw), pos)

    # Make data source w/ time hover
    data = ColumnDataSource(_append_hover(data_plot, multi_day=(extremos[1] - extremos[0]).days > 1))

    # Plot lines of analog sensors
    for s, is_rms, color in zip(COLS_DATA, B_SENSORS_RMS, COLORS_DATA):
        kwargs_p = dict(source=data, alpha=.9, line_width=2, color=color)
        if not is_rms:
            kwargs_p.update(dict(y_range_name=ejes[1]))
        p.line(COL_TS, s, **kwargs_p)

    # Plot patch of main rms & normal analog sensors
    kwargs_patch = dict(color=COLORS_DATA[pos_smean], line_alpha=0, fill_alpha=0.10, y_range_name=ejes[1])
    _plot_patch_series(p, data_plot[COL_MAIN_MEAN], **kwargs_patch)

    kwargs_patch = dict(color=COLORS_DATA[0], line_alpha=0, fill_alpha=0.15)
    _plot_patch_series(p, data_plot[COL_MAIN_RMS], **kwargs_patch)

    # Legend formatting
    # _format_legend_plot(p)
    return p


# TODO Arreglar summary plot
def _plot_bokeh_hourly(data_plot, **fig_kwargs):
    # Bokeh does not work very well!! with timezones:
    data_plot = data_plot.tz_localize(None)

    data_plot['ts_mid'] = data_plot.index + pd.Timedelta('30min')
    data_plot.index += pd.Timedelta('30min')

    y_range = [0, max(1, np.ceil(data_plot[COLS_DATA_KWH[0]].max().max() / ROUND_KWH) * ROUND_KWH)]
    minmax_ejes = [y_range, [0, max(500, int(np.ceil(data_plot[COLS_DATA_KWH[1]].max().max() / ROUND_W) * ROUND_W))]]
    extremos = _get_extremos_dia(data_plot)
    ejes = COLS_DATA_KWH[:2]

    # Figure
    p = _get_figure_plot(extremos, minmax_ejes[0], **fig_kwargs)

    tooltip_rows = TOOLTIP_ROWS
    for c, color, unit, label, fmt in zip(COLS_DATA_KWH, COLORS_DATA_KWH,
                                          UNITS_DATA_KWH, LABELS_DATA_KWH, FMT_TOOLTIP_DATA_KWH):
        tooltip_rows += HTML_TROW.format('{}:'.format(label), '@{}{} {}'.format(c, fmt, unit), color)
    p.add_tools(HoverTool(tooltips='<div><table>{}</table></div>'.format(tooltip_rows)))

    # Axis formatting
    _format_axis_plot(p, COLORS_DATA_KWH[0], LABELS_DATA_KWH[0], UNITS_DATA_KWH[0])
    if len(ejes) > 1:
        positions = (['right', 'left'] * 2)[:len(ejes[1:])]
        for extra_eje, pos, minmax, color, label, unit in zip(ejes[1:], positions, minmax_ejes[1:], COLORS_DATA_KWH[1:],
                                                              LABELS_DATA_KWH[1:], UNITS_DATA_KWH[1:]):
            p.extra_y_ranges[extra_eje] = Range1d(*minmax)
            axkw = dict(y_range_name=extra_eje,
                        axis_label='{} ({})'.format(label, unit),
                        axis_label_text_font_size="10pt", axis_line_color=color,
                        major_label_text_color=color, axis_label_text_color=color,
                        major_tick_line_color=color, minor_tick_line_color=color,
                        major_tick_in=5, major_tick_out=0, minor_tick_in=3, minor_tick_out=0)
            p.add_layout(LinearAxis(**axkw), pos)

    # Make data source w/ time hover
    data = ColumnDataSource(_append_hover(data_plot, delta_min=-30, multi_day=(extremos[1] - extremos[0]).days > 1))

    # Plot lines
    kwargs_kwh = dict(source=data, x=COL_TS, width=3600000, bottom=0, top=COLS_DATA_KWH[0], legend=LABELS_DATA_KWH[0],
                      fill_alpha=.7, line_alpha=.9, line_width=1, line_join='round', color=COLORS_DATA_KWH[0])
    p.vbar(**kwargs_kwh)

    kwargs_l = dict(source=data, alpha=.95, line_width=1.5, line_join='bevel', color=COLORS_DATA_KWH[1],
                    y_range_name=COLS_DATA_KWH[1], legend=LABELS_DATA_KWH[1])
    p.line(COL_TS, COLS_DATA_KWH[1], **kwargs_l)
    kwargs_l.update(color=COLORS_DATA_KWH[2], legend=LABELS_DATA_KWH[2])
    p.line(COL_TS, COLS_DATA_KWH[2], **kwargs_l)

    # Legend formatting
    _format_legend_plot(p)
    return p


def html_plot_buffer_bokeh(data_plot, is_kwh_plot=False, **fig_kwargs):
    """
    Given a pandas DataFrame (or a list of df's), returns the html components for rendering the graph.
    :return script, divs, bokeh.__version__
    """
    if is_kwh_plot:
        return _return_html_comps([_plot_bokeh_hourly(data_plot, **fig_kwargs)])
    else:
        return _return_html_comps([_plot_bokeh_raw(data_plot, **fig_kwargs)])



# # -*- coding: utf-8 -*-
# from pydrive import ...
# from enerpi.base import CONFIG, log, async_task

# TODO GDRIVE SYNC
# RECIPIENT = 'eugenio.panadero@gmail.com'
# EMAIL_ACCOUNT = 'enerpi.bot@gmail.com'
# GMAIL_APP_PASSWORD = 'qkdspbhmxouzrkvv'



# -*- coding: utf-8 -*-
import os
# noinspection PyUnresolvedReferences
from enerpiweb import app as application, SERVER_FILE_LOGGING, LOGGING_LEVEL_SERVER
from enerpiweb.command_enerpiweb import main, check_resource_files
from enerpi.base import set_logging_conf


# Para enviar el output a /var/log/apache2/error.log, utilizado para debug
# import sys
# sys.stdout = sys.stderr

# Establecemos logging
check_resource_files(os.path.dirname(SERVER_FILE_LOGGING), verbose=True)
set_logging_conf(SERVER_FILE_LOGGING, level=LOGGING_LEVEL_SERVER, verbose=True, with_initial_log=True)
# logging.debug('->LOG Estableciendo LOGFILE en {}'.format(SERVER_FILE_LOGGING))


if __name__ == '__main__':
    # Para ejecutar a mano
    main()



# -*- coding: utf-8 -*-
import datetime as dt
import os
import re
import sys
from enerpi import PRETTY_NAME, DESCRIPTION
from enerpi.base import (BASE_PATH, CONFIG, SENSORS, DATA_PATH, CONFIG_FILENAME, show_pi_temperature,
                         FILE_LOGGING, LOGGING_LEVEL, set_logging_conf, log)


# Config:
DEFAULT_IMG_MASK = CONFIG.get('ENERPI_DATA', 'DEFAULT_IMG_MASK', fallback='enerpi_plot_{:%Y%m%d_%H%M}.png')
IMG_TILES_BASEPATH = os.path.join(BASE_PATH, '..', 'enerpiweb', 'static', 'img', 'generated')
UDP_PORT = CONFIG.getint('BROADCAST', 'UDP_PORT', fallback=57775)
HDF_STORE = CONFIG.get('ENERPI_DATA', 'HDF_STORE')


def _enerpi_arguments():
    """
    CLI Parser
    """
    import argparse

    p = argparse.ArgumentParser(description="\033[1m\033[5m\033[32m{}\033[0m\n{}\n\n".format(PRETTY_NAME, DESCRIPTION),
                                epilog='\033[34m\n*** By default, ENERPI starts as receiver (-r) ***\n' +
                                       '\033[0m', formatter_class=argparse.RawTextHelpFormatter)
    g_m = p.add_argument_group(title='☆  \033[1m\033[4mENERPI Working Mode\033[24m',
                               description='→  Choose working mode between RECEIVER / SENDER')
    g_m.add_argument('-e', '--enerpi', action='store_true', help='⚡  SET ENERPI LOGGER & BROADCAST MODE')
    g_m.add_argument('-r', '--receive', action='store_true', help='⚡  SET Broadcast Receiver mode (by default)')
    g_m.add_argument('--port', '--receiver-port', type=int, action='store', default=UDP_PORT, metavar='XX',
                     help='⚡  SET Broadcast Receiver PORT')
    g_m.add_argument('-d', '--demo', action='store_true', help='☮ SET Demo Mode (broadcast random values)')
    g_m.add_argument('--timeout', action='store', nargs='?', type=int, metavar='∆T', const=60,
                     help='⚡  SET Timeout to finish execution automatically')
    g_m.add_argument('--raw', type=int, action='store', nargs='?', const=5, metavar='∆T',
                     help='☮ SET RAW Data Mode (adquire all samples)')
    g_m.add_argument('--config', action='store_true', help='⚒  Shows configuration in INI file')
    g_m.add_argument('--install', action='store_true', help='⚒  Install CRON task for exec ENERPI LOGGER as daemon')
    g_m.add_argument('--uninstall', action='store_true', help='⚒  Delete all CRON tasks from ENERPI')

    g_p = p.add_argument_group(title='︎ℹ️  \033[4mQUERY & REPORT DATA\033[24m')
    filter_24h = (dt.datetime.now().replace(microsecond=0) - dt.timedelta(hours=24)).strftime('%Y-%m-%d %H:%M:%S')
    g_p.add_argument('-f', '--filter', action='store', nargs='?', metavar='TS', const=filter_24h,
                     help='✂ Query the HDF Store with pandas-like slicing:'
                          '\n     "2016-01-07 :: 2016-02-01 04:00" --> df.loc["2016-01-07":"2016-02-01 04:00"]'
                          '\n     \t(Pay atention to the double "::"!!)'
                          '\n     · By default, "-f" filters data from 24h ago (.loc[{}:]).\n\n'.format(filter_24h))
    default_img_nomask = DEFAULT_IMG_MASK.replace('{', '{{').replace('}', '}}').replace('%', '%%')
    help_plot = '''⎙  Plot & save image with matplotlib in any compatible format.
     · If not specified, PNG file is generated with MASK:\n           "{}" using datetime data limits.
     · If only specifying image format, default mask is used with the desired format.
     · If image path is passed, the initial (and final, optionally) timestamps of filtered data
     can be used with formatting masks, like:
         "/path/to/image/image_{{:%%c}}_{{:%%H%%M}}.pdf" or "report_{{:%%d%%m%%y}}.svg".'''.format(default_img_nomask)
    g_p.add_argument('-p', '--plot', action='store', metavar='IM', nargs='?', const=DEFAULT_IMG_MASK, help=help_plot)
    g_p.add_argument('-pt', '--plot_tiles', action='store_true', help='⎙  Generate SVG Tiles for enerpiWeb.')

    g_st = p.add_argument_group(title='⚙  \033[4mHDF Store Options\033[24m')
    g_st.add_argument('--store', action='store', metavar='ST', default=HDF_STORE,
                      help='✏️  Set the .h5 file where save the HDF store.\n     Default: "{}"'.format(HDF_STORE))
    g_st.add_argument('--backup', action='store', metavar='BKP', help='☔ Backup ALL data in CSV format')
    g_st.add_argument('--clearlog', action='store_true', help='⚠ Delete the LOG FILE at: "{}"'.format(FILE_LOGGING))
    g_st.add_argument('-i', '--info', action='store_true', help='︎ℹ Show data info')
    g_st.add_argument('--last', action='store_true', help='︎ℹ Show last saved data')

    g_d = p.add_argument_group(title='☕  \033[4mDEBUG Options\033[24m')
    g_d.add_argument('--temps', action='store_true', help='♨ Show RPI temperatures (CPU + GPU)')
    g_d.add_argument('-l', '--log', action='store_true', help='☕ Show LOG FILE')
    g_d.add_argument('-s', '--silent', action='store_true', help='‼ Silent mode (Verbose mode ON BY DEFAULT in CLI)')

    g_ts = p.add_argument_group(title='⚒  \033[4mCurrent Meter Sampling Configuration\033[24m')
    g_ts.add_argument('-T', '--delta', type=float, action='store', default=SENSORS.delta_sec_data, metavar='∆T',
                      help='⌚  Set Ts sampling (to database & broadcast), in seconds. Default ∆T: {} s'
                      .format(SENSORS.delta_sec_data))
    g_ts.add_argument('-ts', type=int, action='store', default=SENSORS.ts_data_ms, metavar='∆T',
                      help='⏱  Set Ts raw sampling, in ms. Default ∆T_s: {} ms'.format(SENSORS.ts_data_ms))
    g_ts.add_argument('-w', '--window', type=float, action='store', default=SENSORS.rms_roll_window_sec, metavar='∆T',
                      help='⚖  Set window width in seconds for instant RMS calculation. Default ∆T_w: {} s'
                      .format(SENSORS.rms_roll_window_sec))

    return p.parse_args()


def make_cron_command_task_daemon():
    """
    CRON periodic task for exec ENERPI LOGGER as daemon at every boot
    Example command:
    */15 * * * * sudo -u www-data /home/pi/PYTHON/py35/bin/python
        /home/pi/PYTHON/py35/lib/python3.5/site-packages/enerpiweb/mule_rscgen.py -o

    :return: :str: cron_command
    """
    # cmd_logger = '@reboot sudo -u {user_logger} {python_pathbin}/enerpi-daemon start'
    cmd_logger = 'sudo -u {user_logger} {python_pathbin}/enerpi-daemon start'
    local_params = dict(user_logger=CONFIG.get('ENERPI_DATA', 'USER_LOGGER', fallback='pi'),
                        python_pathbin=os.path.dirname(sys.executable))
    return cmd_logger.format(**local_params)


def _check_store_relpath(path_st):
    if os.path.pathsep not in path_st:
        path_st = os.path.join(DATA_PATH, path_st)
    else:
        path_st = os.path.abspath(path_st)
    if not os.path.splitext(path_st)[1]:
        path_st += '.h5'
    existe_st = os.path.exists(path_st)
    if not existe_st:
        log('HDF Store not found at "{}"'.format(path_st), 'warn', True)
    return existe_st, path_st


def enerpi_main_cli(test_mode=False):
    """
    Uso de ENERPI desde CLI

    enerpi -h para mostrar las diferentes opciones

    """
    # Init CLI
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import pandas as pd
    pd.set_option('display.width', 200)

    # CLI Arguments
    args = _enerpi_arguments()
    verbose = not args.silent

    # CONTROL LOGIC
    # Shows RPI Temps
    timer_temps = show_pi_temperature(args.temps, 3)

    if args.install or args.uninstall:
        from enerpi.config.crontasks import set_command_on_reboot, clear_cron_commands
        # INSTALL / UNINSTALL CRON TASKS & KEY
        cmd_logger = make_cron_command_task_daemon()
        if args.install:
            # Logging configuration
            set_logging_conf(FILE_LOGGING, LOGGING_LEVEL, True)

            log('** Installing CRON task for start logger at reboot:\n"{}"'.format(cmd_logger), 'ok', True, False)
            set_command_on_reboot(cmd_logger, verbose=verbose)
            try:
                os.chmod(DATA_PATH, 0o777)
                [os.chmod(os.path.join(base, f), 0o777)
                 for base, dirs, files in os.walk(DATA_PATH) for f in files + dirs]
            except PermissionError:
                log("Can't set 777 permissions on {0}/* files...\nDo it manually, please: 'sudo chmod 777 -R {0}'"
                    .format(DATA_PATH), 'warning', True, False)
        else:
            log('** Deleting CRON task for start logger at reboot:\n"{}"'.format(cmd_logger), 'warn', True, False)
            clear_cron_commands([cmd_logger], verbose=verbose)
    elif (args.enerpi or args.info or args.backup or args.config or args.raw or
            args.last or args.clearlog or args.filter or args.plot or args.plot_tiles):
        # Logging configuration
        set_logging_conf(FILE_LOGGING, LOGGING_LEVEL, True)

        # Shows INI config
        if args.config:
            log('ENERPI Configuration (from INI file in "{}"):'
                .format(os.path.join(DATA_PATH, CONFIG_FILENAME)), 'ok', True, False)
            for s in CONFIG.sections():
                log('* Section {}:'.format(s), 'info', True, False)
                for opt in CONFIG.options(s):
                    log('{:27} -->\t{}'.format(opt.upper(), CONFIG.get(s, opt)), 'debug', True, False)
            log('*' * 80 + '\n', 'ok', True, False)

        # Delete LOG File
        if args.clearlog:
            from enerpi.database import delete_log_file

            delete_log_file(FILE_LOGGING, verbose=verbose)
        # Data Store Config
        _existe_st, path_st = _check_store_relpath(args.store)

        # Starts ENERPI Logger
        if args.enerpi:
            from enerpi.enerpimeter import enerpi_logger

            enerpi_logger(is_demo=False, verbose=verbose, path_st=path_st, delta_sampling=args.delta,
                          roll_time=args.window, sampling_ms=args.ts, timeout=args.timeout)
        elif args.backup:
            from enerpi.database import init_catalog
            # Export data to CSV:
            catalog = init_catalog(sensors=SENSORS, raw_file=path_st, check_integrity=False)
            all_data = catalog.export(args.backup)
            log('ALL DATA:\n{}'.format(all_data), 'ok', True, False)
        elif args.raw:
            from enerpi.enerpimeter import enerpi_raw_data

            # Raw mode
            delta_secs = args.raw
            raw_data = enerpi_raw_data(path_st.replace('.h5', '_raw_sample.h5'), delta_secs=delta_secs,
                                       roll_time=args.window, sampling_ms=args.ts, verbose=verbose)
            t0, tf = raw_data.index[0], raw_data.index[-1]
            log('Showing RAW DATA for {} seconds ({} samples, {:.2f} sps)\n** Real data: from {} to {} --> {:.2f} sps'
                .format(delta_secs, len(raw_data), len(raw_data) / delta_secs,
                        t0, tf, len(raw_data) / (tf-t0).total_seconds()), 'info', verbose, False)
            # TODO revisar config X11 + ssh -X para plot en display local
            raw_data.plot(lw=.5, figsize=(16, 10))
            plt.show()
        # Shows database info
        elif args.info or args.filter or args.plot or args.plot_tiles:
            from enerpi.database import init_catalog, show_info_data

            catalog = init_catalog(sensors=SENSORS, raw_file=path_st, check_integrity=False)
            if args.plot_tiles:
                from enerpiplot.enerplot import gen_svg_tiles

                ok = gen_svg_tiles(IMG_TILES_BASEPATH, catalog)
                if ok:
                    log('SVG Tiles generated!', 'ok', verbose, True)
                else:
                    log('No generation of SVG Tiles!', 'error', verbose, True)
            else:
                if args.filter:
                    loc_data = args.filter.split('::')
                    if len(loc_data) > 1:
                        if len(loc_data[0]) > 0:
                            data, consumption = catalog.get(start=loc_data[0], end=loc_data[1], with_summary=True)
                        else:
                            data, consumption = catalog.get(end=loc_data[1], with_summary=True)
                    else:
                        last_hours = re.findall('(\d{1,5})h', loc_data[0], flags=re.IGNORECASE)
                        if last_hours:
                            data, consumption = catalog.get(last_hours=int(last_hours[0]), with_summary=True)
                        else:
                            data, consumption = catalog.get(start=loc_data[0], with_summary=True)
                else:
                    data, consumption = catalog.get(start=catalog.min_ts, with_summary=True)
                if args.info and data is not None and not data.empty:
                    show_info_data(data, consumption)
                if (args.plot and data is not None and not data.empty and consumption is not None and
                        not consumption.empty):
                    from enerpiplot.enerplot import plot_power_consumption_hourly

                    img_name = plot_power_consumption_hourly(data.power, consumption.kWh, ldr=data.ldr,
                                                             rs_potencia=None, rm_potencia=60, savefig=args.plot)
                    log('Image stored in "{}"'.format(img_name), 'ok', verbose, True)
        # Shows database info
        else:  # Shows last 10 entries
            from enerpi.database import get_ts_last_save

            last = get_ts_last_save(path_st, get_last_sample=True, verbose=verbose, n=10)
            log('Showing last 10 entries in {}:\n{}'.format(path_st, last), 'info', verbose, False)
    # Shows & extract info from LOG File
    elif args.log:
        from enerpi.database import extract_log_file

        data_log = extract_log_file(FILE_LOGGING, extract_temps=True, verbose=verbose)
        try:
            df_temps = data_log[data_log.temp.notnull()].drop(['tipo', 'msg', 'debug_send'], axis=1).dropna(axis=1)
            if len(set(df_temps['exec'])) == 1:
                df_temps = df_temps.drop(['exec'], axis=1)
            path_csv = os.path.join(DATA_PATH, 'debug_rpitemps.csv')
            if not df_temps.empty:
                df_temps.to_csv(path_csv)
                print('*** Saved temperature data extracted from LOG in {}'.format(path_csv))
        except AttributeError:
            print('No se encuentran datos de Tª de RPI en el LOG')
    # Demo sender
    elif args.demo:
        from enerpi.enerpimeter import enerpi_logger

        set_logging_conf(FILE_LOGGING + '_demo.log', LOGGING_LEVEL, True)
        path_st = os.path.join(DATA_PATH, 'debug_buffer_disk.h5')
        enerpi_logger(is_demo=True, verbose=verbose, path_st=path_st, delta_sampling=args.delta,
                      roll_time=args.window, sampling_ms=args.ts, timeout=args.timeout)
    # Receiver
    else:  # elif args.receive:
        from enerpi.enerpimeter import receiver

        log('{}\n   {}'.format(PRETTY_NAME, DESCRIPTION), 'ok', verbose, False)
        receiver(verbose=verbose, timeout=args.timeout, port=args.port)
    if timer_temps is not None:
        log('Stopping RPI TEMPS sensing...', 'debug', True)
        timer_temps.cancel()
    log('Exiting from ENERPI CLI...', 'debug', True)
    if not test_mode:
        sys.exit(0)


if __name__ == '__main__':
    enerpi_main_cli(test_mode=False)



# -*- coding: utf-8 -*-
import cairosvg
from configparser import NoOptionError, NoSectionError
import datetime as dt
import os
from pushbullet import Pushbullet
from time import time
from enerpi.base import CONFIG, log, async_task, RECIPIENT


# Pushbullet API config:
# Pushbullet channel:
# enerpi_notifications
# ENERPI
# ENERPI Current Meter BOT
# https://www.pushbullet.com/channel?tag=enerpi_notifications
CHANNEL = 'enerpi_notifications'
try:
    PUSHBULLET_TOKEN = CONFIG.get('NOTIFY', 'PUSHBULLET_TOKEN')
except (NoOptionError, NoSectionError) as e:
    PUSHBULLET_TOKEN = None
    log('NO Pushbullet config ({} [{}])'.format(e, e.__class__), 'error', False)
PBOBJ = None
if PUSHBULLET_TOKEN is not None:
    PBOBJ = Pushbullet(PUSHBULLET_TOKEN)


def _send_direct_push_notification(title, content, email=RECIPIENT):
    if PUSHBULLET_TOKEN is not None:
        pb = PBOBJ
        pb.push_note(title, content, email=email)
        return True
    log('NO PUSHBULLET NOTIFICATION (token={}), title={}, email={}'.format(PUSHBULLET_TOKEN, title, email),
        'error', True)
    return False


@async_task
def push_enerpi_error(title_err, msg_error):
    """
    Send an error message notification (vía pushbullet API) to the default recipient.
    :param title_err: error title
    :param msg_error: error text content

    """
    return _send_direct_push_notification('ERROR: {}\n(time:{:%H:%M:%S})'.format(title_err, dt.datetime.now()),
                                          msg_error, email=RECIPIENT)


@async_task
def push_enerpi_channel_msg(title, msg, channel_tag=CHANNEL):
    """
    Send a text notification (vía pushbullet API) to the ENERPI pushbullet channel (timestamped)

    :param title:
    :param msg:
    :param channel_tag: ENERPI CHANNEL

    """
    content = '{}\n(time:{:%H:%M:%S})'.format(msg, dt.datetime.now())
    if PUSHBULLET_TOKEN is not None:
        # pb = Pushbullet(PUSHBULLET_TOKEN)
        pb = PBOBJ
        try:
            ch = list(filter(lambda x: x.channel_tag == channel_tag, pb.channels))[0]
        except IndexError:
            ch = None
        if ch:
            ch.push_note(title, content)
            return True
    log('NO PUSHBULLET NOTIFICATION (token={}), title={}, channel={}'.format(PUSHBULLET_TOKEN, title, channel_tag),
        'error', True)
    return False


@async_task
def push_enerpi_file(path_file, title=None, email=RECIPIENT, channel_tag=None,
                     svg_conversion_format='png', verbose=False):
    """
    Send a file with pushbullet (to a recipient or to a channel

    :param path_file:
    :param title:
    :param email:
    :param channel_tag:
    :param svg_conversion_format:
    :param verbose:

    """
    if not os.path.exists(path_file):
        log('Trying to push a non existent file! ->'.format(path_file), 'error', verbose)
        return False
    elif PUSHBULLET_TOKEN is None:
        log('NO PUSHBULLET FILE PUSH (token={}), title={}, email={}, channel={}'
            .format(PUSHBULLET_TOKEN, title, email, channel_tag), 'error', True)
        return False
    tic = time()
    name_file = os.path.basename(path_file)
    # SVG to PNG if needed:
    if path_file.endswith('.svg') and svg_conversion_format:
        name_file = name_file[:-3] + svg_conversion_format
        log('SVG to {} Conversion: {} --> {}'
            .format(svg_conversion_format.upper(), path_file, name_file), 'warning', verbose)
        with open(path_file, 'r') as f:
            svg_content = f.read()
        # SVG Transformation --> some background
        style = "fill:none;opacity:0;"
        new_style = "fill:#FFFFFF;opacity:0.1;"
        svg_content = svg_content.replace(style, new_style)
        # Cairo conversion
        func_convers = getattr(cairosvg, 'svg2{}'.format(svg_conversion_format.lower()))
        b_content_file = func_convers(bytestring=svg_content.encode(), dpi=300, unsafe=True)
    else:
        with open(path_file, 'rb') as f:
            b_content_file = f.read()
    tic_i = time()
    # pb = Pushbullet(PUSHBULLET_TOKEN)
    pb = PBOBJ
    file_data = pb.upload_file(b_content_file, name_file)
    toc = time()
    log('PushBullet file upload: {} (name_pic= {}). Took {:.3f} sec'
        .format(path_file, name_file, toc - tic), 'info', verbose)
    if title is not None:
        file_data.update({'title': title})
    if email is not None:
        file_data.update({'email': email})
    else:
        try:
            ch = list(filter(lambda x: x.channel_tag == channel_tag, pb.channels))[0]
        except IndexError:
            ch = None
        if ch:
            file_data.update({'channel': ch})
        else:
            log('No se encuentra el canal: "{}". Los disponibles son:\n{}'
                .format(channel_tag, pb.channels), 'error', verbose)
            return False
    # Envía notificación
    pb.push_file(**file_data)
    log('PB_NOTIF TOOK: PROCESS_FILE={:.2f}s; UPLOAD_FILE={:.2f}s; PUSH_FILE={:.2f}s'
        .format(tic_i - tic, toc - tic_i, time() - toc), 'debug', verbose)
    return True



# -*- coding: utf-8 -*-
import configparser
import json
import locale
import logging
import os
import pytz
import shutil
import subprocess
import sys
from threading import Thread, Timer
from time import time, sleep
from enerpi import BASE_PATH, PRETTY_NAME
from enerpi.prettyprinting import print_err, print_red, print_info, print_ok, print_warn, print_yellowb, print_magenta
from enerpi.pitemps import get_cpu_temp, get_gpu_temp


ENCODING = 'UTF-8'
CONFIG_FILENAME = 'config_enerpi.ini'
SENSORS_CONFIG_JSON_FILENAME = 'sensors_enerpi.json'


class EnerpiAnalogSensor(object):
    """
    Object for centralize access to sensor properties, as description, color, units, etc.

    """
    def __init__(self, column_name, description='', channel=0, is_rms=True, unit='W', color='#0CBB43', bias=0,
                 tile_gradient_st='', tile_gradient_end='', icon='bolt'):
        self.name = column_name
        self.description = description
        self.channel = channel
        self.is_rms = is_rms
        self.unit = unit
        self.color = color
        self.bias = bias
        self.icon = icon
        self.tile_gradient_st = tile_gradient_st
        self.tile_gradient_end = tile_gradient_end

    def __repr__(self):
        str_type = 'RMS, ' if self.is_rms else ''
        return ('ANALOG_S: {} ({}{}, ch={}, unit={}, color={}, bias={:.3f})'
                .format(self.name, str_type, self.description, self.channel, self.unit, self.color, self.bias))

    def to_dict(self):
        """
        Return dict with analog sensor properties for jinja2 templating

        :return: dict

        """
        return dict(name=self.name, description=self.description, channel=self.channel, is_rms=self.is_rms,
                    unit=self.unit, color=self.color, bias=self.bias,
                    icon=self.icon, tile_gradient_st=self.tile_gradient_st, tile_gradient_end=self.tile_gradient_end)


def _load_analog_sensors(sensors_json_conf):
    """
    Load list of 'EnerpiAnalogSensor' objects, with the ENERPI sensors user configuration
    Format of json file:
    [ {
        "name": "power",
        "analog_channel": 3,
        "is_rms": true,
        "unit": "W",
        "color": "#0CBB43",
        "tile_gradient_st": [12, 187, 67, 0.83],
        "tile_gradient_end": [55, 245, 119, 0.27],
        "icon": "flash",
        "description": "Main Power"
      }, {...}, ... ]

    :param sensors_json_conf: list of dicts (from JSON config file)
    :return list of analog sensors

    """
    def _get_rms_sensor_bias(sensor):
        if sensor['is_rms']:
            if 'bias' in sensor:
                return sensor['bias']
            # Voltage divisor for sct030 probes:
            mcp3008_dac_prec = 10  # bits
            niveles = 2 ** mcp3008_dac_prec - 1
            bias_current = -(niveles // 2) / niveles
            return bias_current
        return 0.

    analog_s = [EnerpiAnalogSensor(s['name'], s['description'], s['analog_channel'], s['is_rms'], s['unit'], s['color'],
                                   _get_rms_sensor_bias(s), s['tile_gradient_st'], s['tile_gradient_end'], s['icon'])
                for s in sensors_json_conf]  # JSON file sensor order
    return analog_s


class EnerpiSamplerConf(object):
    """
    Object for easy access to data properties, as # of columns, column names, descriptions and units, type of sensors,
    color for plotting, summary info & columns, etc...

    """
    def __init__(self, configparser_obj, sensors_json_conf):
        # Read config INI file
        self._fmt_ts = configparser_obj.get('ENERPI_SAMPLER', 'FMT_TS', fallback='%Y-%m-%d %H:%M:%S.%f')
        self._col_ts = configparser_obj.get('ENERPI_SAMPLER', 'COL_TS', fallback='ts')
        self._ref_column_rms = 'ref'
        self._ref_column_mean = 'ref_n'

        # Get sensors:
        sensors = _load_analog_sensors(sensors_json_conf)
        self._sensors_rms = [s for s in sensors if s.is_rms]
        self._sensors_mean = [s for s in sensors if not s.is_rms]

        # PiSampler parameters (Current meter):
        # Voltaje típico RMS de la instalación a medir. (SÓLO SE ESTIMA P_ACTIVA!!)
        voltaje = CONFIG.getint('ENERPI_SAMPLER', 'VOLTAJE', fallback=236)
        # 30 A para 1 V --> Pinza amperométrica SCT030-030
        a_ref = CONFIG.getfloat('ENERPI_SAMPLER', 'A_REF', fallback=30.)
        # V, V_ref RPI GPIO
        v_ref = CONFIG.getfloat('ENERPI_SAMPLER', 'V_REF', fallback=3.3)
        self.voltaje = voltaje
        self.rms_multiplier = self.voltaje * a_ref * v_ref

        # Sampling:
        self.ts_data_ms = CONFIG.getfloat('ENERPI_SAMPLER', 'TS_DATA_MS', fallback=12)
        # ∆T para el deque donde se acumulan frames
        self.rms_roll_window_sec = CONFIG.getfloat('ENERPI_SAMPLER', 'RMS_ROLL_WINDOW_SEC', fallback=2.)
        s_calc = self.ts_data_ms if self.ts_data_ms > 0 else 8
        self.n_samples_buffer_rms = int(round(self.rms_roll_window_sec * 1000 / s_calc))

        self.delta_sec_data = CONFIG.getfloat('ENERPI_SAMPLER', 'DELTA_SEC_DATA', fallback=1.)
        self.measure_ldr_divisor = CONFIG.getint('ENERPI_SAMPLER', 'MEASURE_LDR_DIVISOR', fallback=10)
        self.n_samples_buffer_mean = self.n_samples_buffer_rms // self.measure_ldr_divisor
        self.TZ = pytz.timezone(CONFIG.get('ENERPI_SAMPLER', 'TZ', fallback='Europe/Madrid'))

    def __repr__(self):
        repr_str = 'ENERPI SAMPLER SENSORS:\n * '
        repr_str += '\n * '.join([str(s) for s in self])
        repr_str += ('\n --SAMPLING: V={} V; ∆T={} s; sampling={} ms, N={}, {}(/{})'
                     .format(self.voltaje, self.delta_sec_data, self.ts_data_ms,
                             self.n_samples_buffer_rms, self.n_samples_buffer_mean, self.measure_ldr_divisor))
        return repr_str

    def __len__(self):
        return len(self._sensors_rms) + len(self._sensors_mean)

    def __getitem__(self, key):
        if type(key) is int:
            if key >= len(self._sensors_rms):
                try:
                    return self._sensors_mean[key - len(self._sensors_rms)]
                except IndexError:
                    raise KeyError('Sensor #{} not present (available sensors: {})'
                                   .format(key, self.columns_sensors[1:]))
            return self._sensors_rms[key]
        else:
            assert type(key) is str
            try:
                int_key = self.columns_sensors.index(key) - 1
                return self.__getitem__(int_key)
            except ValueError:
                raise KeyError("Sensor '{}' not present (available sensors: {})".format(key, self.columns_sensors[1:]))

    def __iter__(self):
        for x in self._sensors_rms + self._sensors_mean:
            yield x

    def to_dict(self):
        """
        Return dict of analog sensors for jinja2 templating

        :return: dict

        """
        d_sensors = {'sensors': [x.to_dict() for x in self._sensors_rms + self._sensors_mean]}
        return d_sensors

    @property
    def main_column(self):
        """
        First (& principal) RMS sensor (Main power)

        :return: :str: name of main column
        """
        return self._sensors_rms[0].name

    @property
    def ts_column(self):
        """
        Name of time-series index

        :return: :str: name of time-series index column
        """
        return self._col_ts

    @property
    def ts_fmt(self):
        """
        String formatting of timestamps:

        :return: :str: fmt
        """
        return self._fmt_ts

    @property
    def ref_rms(self):
        """
        Name of column with # of samples of RMS values

        :return: :str: column name
        """
        return self._ref_column_rms

    @property
    def ref_mean(self):
        """
        Name of column with # of samples of MEAN values

        :return: :str: column name
        """
        return self._ref_column_mean

    @property
    def n_cols_sensors(self):
        """
        # of columns: index column (TS) + RMS sensors + MEAN sensors

        :return: :int: # of columns
        """
        return len(self) + 1

    @property
    def n_cols_sampling(self):
        """
        # of columns: index column (TS) + RMS sensors + MEAN sensors + ref_RMS + ref_MEAN

        :return: :int: # of columns
        """
        return len(self) + 3

    @property
    def columns_sensors(self):
        """
        List of columns names of sensors (+ts): index column (TS) + RMS sensors + MEAN sensors

        :return: :list:
        """
        return [self.ts_column] + [s.name for s in self._sensors_rms + self._sensors_mean]

    @property
    def columns_sensors_rms(self):
        """
        List of columns names of RMS sensors

        :return: :list:
        """
        return [s.name for s in self._sensors_rms]

    @property
    def columns_sensors_mean(self):
        """
        List of columns names of MEAN sensors

        :return: :list:
        """
        return [s.name for s in self._sensors_mean]

    @property
    def columns_sampling(self):
        """
        List of columns names of sampling data index column (TS) + RMS sensors + MEAN sensors + REF_RMS + REF_MEAN

        :return: :list:
        """
        return self.columns_sensors + [self.ref_rms, self.ref_mean]

    def included_columns_sensors(self, dict_sample):
        """
        Return RMS & MEAN columns included in dict sample

        :return: :tuple: tuple with 2 lists (rms & mean included columns)

        """
        cols_rms = [s.name for s in self._sensors_rms if s.name in dict_sample]
        cols_mean = [s.name for s in self._sensors_mean if s.name in dict_sample]
        return cols_rms, cols_mean

    def included_columns_sampling(self, dict_sample):
        """
        Return RMS & MEAN columns (or ref_RMS and ref_MEAN) included in sampling data dict

        :return: :tuple: tuple with 3 lists (rms, mean & ref included columns)

        """
        cols_rms, cols_mean = self.included_columns_sensors(dict_sample)
        cols_ref = [c for c in [self._ref_column_rms, self._ref_column_mean] if c in dict_sample]
        return cols_rms, cols_mean, cols_ref

    def descriptions(self, columns, return_list=True):
        """
        Return column descriptions (as list or as dict c:desc)
        :param columns: list of columns
        :param return_list: :bool: False for dict return
        :return: :list: or :dict:
        """
        def _get_desc(key):
            try:
                return self[key].description
            except KeyError as e:
                if key == self.ts_column:
                    return 'TS'
                elif key == self.ref_rms:
                    return '# samples'
                elif key == self.ref_mean:
                    return '# samples_n'
                else:
                    raise KeyError(e)

        if return_list:
            return [_get_desc(c) for c in columns]
        else:
            return {c: _get_desc(c) for c in columns}


def _makedirs(dest_path):
    try:
        os.makedirs(os.path.dirname(dest_path), exist_ok=True)
        if os.path.exists(dest_path):
            os.utime(dest_path, None)
    except PermissionError as e:
        from time import sleep
        from enerpi.notifier import push_enerpi_error

        msg_error = '_makedirs PermissionError (dest_path = {}). Exception: {}. Exit in 2 secs...'.format(dest_path, e)
        log(msg_error, 'error', True, False)
        push_enerpi_error('ENERPI Fatal error', msg_error, channel_tag='raspberries')
        sleep(2)
        sys.exit(2)


def _funcs_tipo_output(tipo_log):
    """
    Functions for printing and logging based on type.

    :param tipo_log: :enum: error, debug, ok, info, warn, magenta
    :return: print_func, logging_func
    """
    if tipo_log == 'error':
        return print_err, logging.error
    elif tipo_log == 'debug':
        return print_red, logging.debug
    elif tipo_log == 'ok':
        return print_ok, logging.info
    elif tipo_log == 'info':
        return print_info, logging.info
    elif (tipo_log == 'warn') or (tipo_log == 'warning'):
        return print_warn, logging.warning
    elif tipo_log == 'magenta':
        return print_magenta, logging.warning
    else:
        return print_yellowb, logging.debug


def log(msg, tipo, verbose=True, log_msg=True):
    """
    Logging wrapper to log and / or color-print

    :param msg:
    :param tipo:
    :param verbose: :bool:
    :param log_msg: :bool:
    """
    f1, f2 = _funcs_tipo_output(tipo)
    if verbose:
        f1(msg)
    if log_msg:
        f2(msg)


def set_logging_conf(filename, level='DEBUG', verbose=True, with_initial_log=True):
    """Logging configuration"""
    _makedirs(filename)
    # os.makedirs(os.path.dirname(filename), exist_ok=True)
    # try:
    logging.basicConfig(filename=filename, level=level, datefmt='%d/%m/%Y %H:%M:%S',
                        format='%(levelname)s [%(filename)s_%(funcName)s] - %(asctime)s: %(message)s')
    # except PermissionError as e:
    #     log('PermissionError: {}'.format(e), 'error', True, False)
    #     sudo chmod 777 ~/ENERPIDATA/enerpi.log
    #     sys.exit(2)
    if with_initial_log:
        log(PRETTY_NAME, 'ok', verbose)


def show_pi_temperature(log_rpi_temps, ts=3):
    """
    Sensor Raspberry PI temperatures, infinite-loop of logging / printing values

    :param log_rpi_temps: :bool: activate RPI temp logging
    :param ts: :int: ∆T between samples
    """
    def _loop_log_temps():
        while True:
            t_cpu = get_cpu_temp()
            t_gpu = get_gpu_temp()
            if (t_cpu is None) and (t_gpu is None):
                log('NO RPI TEMPS', 'warning', False, True)
                break
            log('Tªs --> {:.1f} / {:.1f} ºC'.format(t_cpu, t_gpu), 'otro', False, True)
            sleep(ts)

    if log_rpi_temps:
        # Shows RPI Temps
        timer_temps = Timer(.5, _loop_log_temps)
        timer_temps.start()
        return timer_temps
    return None


def async_task(f):
    """
    Decorator (wrapper) for execute an async action (threaded)
    :param f:

    """
    def _wrapper(*args, **kwargs):
        thr = Thread(target=f, args=args, kwargs=kwargs)
        thr.start()
    return _wrapper


def timeit(cadena_log, verbose=False, *args_dec):
    """
    Decorator (wrapper) to timeit and log (and print) any function.

    (For debugging / optimize purposes)

    :param cadena_log:
    :param verbose:
    :param args_dec:
    :return:
    """
    def _real_deco(function):
        def _wrapper(*args, **kwargs):
            kwargs_print = {}
            tic = time()
            out = function(*args, **kwargs)
            if verbose:
                print_yellowb(cadena_log.format(*args_dec, **kwargs_print) + ' TOOK: {:.3f} s'.format(time() - tic))
            logging.debug(cadena_log.format(*args_dec, **kwargs_print) + ' TOOK: {:.3f} s'.format(time() - tic))
            return out
        return _wrapper
    return _real_deco


def get_lines_file(filename, tail=None, reverse=False):
    """
    Read (or tail) a text file.

    :param filename:
    :param tail:
    :param reverse:
    :return:
    """
    if os.path.exists(filename):
        if os.path.isfile(filename):
            try:
                if tail is not None:
                    output = subprocess.check_output(['/usr/bin/tail', '-n', '{}'.format(int(tail)), filename])
                    lines = output.decode().split('\n')
                else:
                    with open(filename, 'r', encoding=ENCODING) as file:
                        lines = file.read().split('\n')
                if len(lines) > 0 and lines[-1] == '':
                    lines = lines[:-1]
                if reverse:
                    return list(reversed(lines))
                return lines
            except Exception as e:
                return ['ERROR Reading {}: "{}" [{}]'.format(filename, e, e.__class__)]
        else:
            return ['{} is not a file!!'.format(filename)]
    return ['Path not found: "{}"'.format(filename)]


def check_resource_files(dest_path, origin_path=None, verbose=True):
    """
    Check needed files and directories in DATA_PATH. Init if needed (1º exec).

    :param dest_path:
    :param origin_path:
    :param verbose:

    """
    if not os.path.exists(dest_path):
        _makedirs(dest_path)
        if origin_path is not None:
            origin_path = os.path.abspath(origin_path)
            if os.path.isfile(origin_path):
                log('-> Copying resource file from "{}" to "{}"'.format(origin_path, dest_path), 'info', verbose)
                shutil.copy(origin_path, dest_path)
            else:
                log('-> Replicating tree from "{}" to "{}"'.format(origin_path, dest_path), 'info', verbose)
                shutil.copytree(origin_path, dest_path)
        log('** check_resource_files OK', 'debug', verbose)
    elif oct(os.stat(dest_path).st_mode)[-3:] != '777':
        log('Permission trouble in "{}" --> {}'
            .format(dest_path, oct(os.stat(dest_path).st_mode)[-3:]), 'warning', verbose)
        try:
            os.chmod(dest_path, 0o777)
        except PermissionError:
            log('Permission error in "{}" --> {}. Cant set 777'
                .format(dest_path, oct(os.stat(dest_path).st_mode)[-3:]), 'error', verbose)


def get_config_enerpi():
    """
    Loads or generates ini file for ENERPI (& ENERPIweb) configuration.

    1) Looks for variable path DATA_PATH in file .../enerpi/config/.enerpi_data_path
    2) Tries to load 'config_enerpi.ini' from DATA_PATH, as user custom config.
    2) Tries to load 'sensors_enerpi.json' from DATA_PATH, as user custom settings for analog sensors.
    3) If not present, generates it copying the default configuration.

    :return: :str: data_path, :configparser: loaded object, :dict:, enerpi sensors configuration
    """

    # Load DATA_PATH:
    dir_config = os.path.join(BASE_PATH, 'config')
    path_default_datapath = os.path.join(dir_config, '.enerpi_data_path')
    try:
        with open(path_default_datapath, 'r', encoding=ENCODING) as f:
            raw = f.read()
        data_path = raw.split('\n')[0]
        if data_path != os.path.expanduser(data_path):  # Hay '~', se expande el usuario y se graba abspath
            data_path = os.path.expanduser(data_path)
            log('''Sobreescritura del archivo "{}",
almacenando la ruta absoluta a la instalación de ENERPI
    -> DATA_PATH := {}
** Para mover la instalación de ENERPI a otro lugar, edite directamente este fichero
   (y mueva manualmente la carpeta DATA_PATH)'''.format(path_default_datapath, data_path), 'info', True, True)
            with open(path_default_datapath, 'w', encoding=ENCODING) as f:
                f.write(data_path)
    except Exception as e:
        log('ENERPI LOAD CONFIG ERROR at "{}" --> {} [{}]'
            .format(path_default_datapath, e, e.__class__), 'error', True, True)
        data_path = os.path.expanduser('~/ENERPIDATA')

    # Checks paths & re-gen configuration if not existent
    check_resource_files(data_path)

    # Config parser
    path_file_config = os.path.join(data_path, CONFIG_FILENAME)
    if not os.path.exists(path_file_config):
        check_resource_files(path_file_config)
        log('** Instalando fichero de configuración en: "{}"'.format(path_file_config), 'info', True, True)
        shutil.copy(os.path.join(dir_config, 'default_config_enerpi.ini'), path_file_config)
    configp = configparser.RawConfigParser()
    try:
        configp.read(path_file_config, encoding=ENCODING)
    except Exception as e:
        log('Error loading configuration INI file in "{}". Exception {} [{}]. Using defaults...'
            .format(path_file_config, e, e.__class__), 'error', True, True)
        configp.read(os.path.join(dir_config, 'default_config_enerpi.ini'), encoding=ENCODING)

    # Sensors
    path_file_sensors_config = os.path.join(data_path, SENSORS_CONFIG_JSON_FILENAME)
    if not os.path.exists(path_file_sensors_config):
        check_resource_files(path_file_sensors_config)
        log('** Instalando fichero de configuración en: "{}"'.format(path_file_sensors_config), 'info', True, True)
        shutil.copy(os.path.join(dir_config, 'default_sensors_enerpi.json'), path_file_sensors_config)
    try:
        config_s = json.load(open(path_file_sensors_config, encoding=ENCODING))
    except Exception as e:
        log('Error loading sensors JSON configuration file in "{}". Exception {} [{}]. Using defaults...'
            .format(path_file_sensors_config, e, e.__class__), 'error', True, True)
        config_s = json.load(open(os.path.join(dir_config, 'default_sensors_enerpi.json'), encoding=ENCODING))
    return data_path, configp, config_s


# TODO Revisión reload_config
def reload_config():
    global SENSORS, DATA_PATH, CONFIG, SENSORS_THEME

    DATA_PATH, CONFIG, SENSORS_THEME = get_config_enerpi()
    SENSORS = EnerpiSamplerConf(CONFIG, SENSORS_THEME)


# Loads configuration
DATA_PATH, CONFIG, SENSORS_THEME = get_config_enerpi()

# Appends ENERPI gmail account: (Hardcoded)
GMAIL_ACCOUNT = 'enerpi.bot@gmail.com'
GMAIL_APP_PASSWORD = 'qkdspbhmxouzrkvv'

# Admin email for reports & nofifications:
RECIPIENT = CONFIG.get('NOTIFY', 'RECIPIENT', fallback='eugenio.panadero@gmail.com')

# Set Locale
CUSTOM_LOCALE = CONFIG.get('ENERPI_SAMPLER', 'LOCALE', fallback='{}.{}'.format(*locale.getlocale()))
locale.setlocale(locale.LC_ALL, CUSTOM_LOCALE)

# ANALOG SENSORS WITH MCP3008 (Rasp.io Analog Zero)
SENSORS = EnerpiSamplerConf(CONFIG, SENSORS_THEME)

# Logging files & other common paths:
FILE_LOGGING = os.path.join(DATA_PATH, CONFIG.get('ENERPI_DATA', 'FILE_LOGGING', fallback='enerpi.log'))
LOGGING_LEVEL = CONFIG.get('ENERPI_DATA', 'LOGGING_LEVEL', fallback='DEBUG')

STATIC_PATH = os.path.join(DATA_PATH, CONFIG.get('ENERPI_WEBSERVER', 'STATIC_PATH', fallback='WWW'))
SERVER_FILE_LOGGING_RSCGEN = os.path.join(STATIC_PATH, 'enerpiweb_rscgen.log')
NGINX_CONFIG_FILE = 'enerpiweb_nginx.conf'
UWSGI_CONFIG_FILE = 'enerpiweb_uwsgi.ini'
IMG_TILES_BASEPATH = os.path.join(STATIC_PATH, 'img', 'generated')
DAEMON_STDOUT = '/tmp/enerpi_out.txt'
DAEMON_STDERR = '/tmp/enerpi_err.txt'
DAEMON_PIDFILE = '/tmp/enerpilogger.pid'
INDEX_DATA_CATALOG = 'data_catalog.csv'



# -*- coding: utf-8 -*-
from flask import request, redirect, url_for, render_template, send_file, abort
from flask_mail import Message
import json
import os
from enerpi.base import (get_lines_file, log, DATA_PATH, FILE_LOGGING, SERVER_FILE_LOGGING_RSCGEN,
                         RECIPIENT, SENSORS, INDEX_DATA_CATALOG,
                         UWSGI_CONFIG_FILE, DAEMON_STDOUT, DAEMON_STDERR)
from enerpi.editconf import web_config_edition_data, check_uploaded_config_file, ENERPI_CONFIG_FILES
from enerpi.api import enerpi_data_catalog
from enerpiweb import app, auto, mail, STATIC_PATH, GMAIL_ACCOUNT, SERVER_FILE_LOGGING
from enerpiweb.forms import FileForm


def _get_filepath_from_file_id(file_id):
    # Interesting files / logs to show/edit/download/upload:
    if file_id in ENERPI_CONFIG_FILES:
        return True, False, os.path.join(DATA_PATH, ENERPI_CONFIG_FILES[file_id]['filename'])
    elif 'flask' == file_id:
        is_logfile, filename = True, SERVER_FILE_LOGGING
    elif 'rsc' == file_id:
        is_logfile, filename = True, SERVER_FILE_LOGGING_RSCGEN
    elif 'nginx_err' == file_id:
        is_logfile, filename = True, '/var/log/nginx/error.log'
    elif 'nginx' == file_id:
        is_logfile, filename = True, '/var/log/nginx/access.log'
    elif 'enerpi' == file_id:
        is_logfile, filename = True, FILE_LOGGING
    elif 'uwsgi' == file_id:
        is_logfile, filename = True, '/var/log/uwsgi/{}.log'.format(os.path.splitext(UWSGI_CONFIG_FILE)[0])
    elif 'daemon_out' == file_id:
        is_logfile, filename = True, DAEMON_STDOUT
    elif 'daemon_err' == file_id:
        is_logfile, filename = True, DAEMON_STDERR
    else:  # Fichero derivado del catálogo
        cat = enerpi_data_catalog(check_integrity=False)
        if 'raw_store' == file_id:
            is_logfile, filename = False, os.path.join(cat.base_path, cat.raw_store)
        elif 'catalog' == file_id:
            is_logfile, filename = False, os.path.join(cat.base_path, cat.catalog_file)
        else:
            log('FILE_ID No reconocido: {}'.format(file_id), 'error', False)
            is_logfile, filename = False, SERVER_FILE_LOGGING
            return False, is_logfile, filename
    return True, is_logfile, filename


#############################
# ROUTES for file handling
#############################
@app.route('/api/hdfstores/<relpath_store>', methods=['GET'])
@auto.doc()
def download_hdfstore_file(relpath_store=None):
    """
    Devuelve el fichero HDFStore del catálogo de ENERPI pasado como ruta relativa (nombre de fichero .h5).
    * File Download *
    :param relpath_store:
    """
    cat = enerpi_data_catalog(check_integrity=False)
    path_file = cat.get_path_hdf_store_binaries(relpath_store)
    log('download_hdfstore_file with path_file: "{}", relpath: "{}"'.format(path_file, relpath_store), 'debug', False)
    if (path_file is not None) and os.path.exists(path_file):
        if 'as_attachment' in request.args:
            return send_file(path_file, as_attachment=True, attachment_filename=os.path.basename(path_file))
        return send_file(path_file, as_attachment=False)
    return abort(404)


@app.route('/api/filedownload/<file_id>', methods=['GET'])
@auto.doc()
def download_file(file_id):
    """
    * File Download *
    :param file_id:
    """
    ok, _, filename = _get_filepath_from_file_id(file_id)
    if ok:
        if os.path.exists(filename):
            if 'as_attachment' in request.args:
                return send_file(filename, as_attachment=True, attachment_filename=os.path.basename(filename))
            return send_file(filename, as_attachment=False)
        else:
            msg = json.dumps({'alert_type': 'danger',
                              'texto_alerta': 'El archivo "{}" ({}) no existe!'.format(filename, file_id)})
            return redirect(url_for('control', alerta=msg))
    return abort(404)


@app.route('/api/filedownload/debug/', methods=['POST'])
@auto.doc()
def download_file_debug():
    """
    * File Download *
    Post request with relative file path under DATA_PATH (debugging purposes)

    """
    form = FileForm()
    if form.validate_on_submit():
        # relpath = relpath.split(':')
        relpath = form.pathfile
        # abspath = os.path.join(DATA_PATH, *relpath)
        abspath = os.path.join(DATA_PATH, relpath)
        if os.path.exists(abspath):
            return send_file(abspath, as_attachment=True, attachment_filename=os.path.basename(abspath))
    return abort(404)


@app.route('/api/showfile', methods=['GET'])
@app.route('/api/showfile/<file>', methods=['GET'])
@auto.doc()
def showfile(file='flask'):
    """
    Página de vista de fichero de texto, con orden ascendente / descendente y/o nº de últimas líneas ('tail' de archivo)

    :param file: file_id to show

    """
    ok, is_logfile, filename = _get_filepath_from_file_id(file)
    if ok:
        delete = request.args.get('delete', '')
        reverse = request.args.get('reverse', False)
        tail_lines = request.args.get('tail', None)
        alerta = request.args.get('alerta', '')
        if alerta:
            alerta = json.loads(alerta)
        if not alerta and delete and is_logfile:
            with open(filename, 'w') as f:
                f.close()
            cad_delete = 'LOGFILE {} DELETED'.format(filename.upper())
            log(cad_delete, 'warn', False)
            return redirect(url_for('showfile', file=file,
                                    alerta=json.dumps({'alert_type': 'warning', 'texto_alerta': cad_delete})))
        data = get_lines_file(filename, tail=tail_lines, reverse=reverse)
        return render_template('text_file.html', titulo='LOG File:' if is_logfile else 'CONFIG File:', file_id=file,
                               subtitulo='<strong>{}</strong>'.format(filename), is_logfile=is_logfile,
                               file_content=data, filename=os.path.basename(filename), alerta=alerta)
    return abort(404)


@app.route('/api/editconfig/', methods=['GET'])
@app.route('/api/editconfig/<file>', methods=['GET', 'POST'])
@auto.doc()
def editfile(file='config'):
    """
    Configuration editor, for INI file, JSON sensors file & encripting key

    :param file: file_id to edit
    """
    ok, is_logfile, filename = _get_filepath_from_file_id(file)
    if ok and (file in ENERPI_CONFIG_FILES):
        alerta = request.args.get('alerta', '')
        if alerta:
            alerta = json.loads(alerta)
        extra_links = [(ENERPI_CONFIG_FILES[f_id]['text_button'], url_for('editfile', file=f_id))
                       for f_id in sorted(ENERPI_CONFIG_FILES.keys(), key=lambda x: ENERPI_CONFIG_FILES[x]['order'])]
        show_switch_comments = ENERPI_CONFIG_FILES[file]['show_switch_comments']
        if not show_switch_comments:
            without_comments = True
        else:
            without_comments = request.args.get('without_comments', 'False').lower() == 'true'
        d_edit = request.form if request.method == 'POST' else None
        alerta_edit, lines_config, config_data = web_config_edition_data(file, filename, d_edition_form=d_edit)
        if not alerta:
            alerta = alerta_edit
        elif alerta_edit is not None:
            alerta.update(alerta_edit)
        return render_template('edit_text_file.html', titulo='CONFIG EDITOR', file_id=file,
                               show_switch_comments=show_switch_comments, with_comments=not without_comments,
                               abspath=filename, dict_config_content=config_data, file_lines=lines_config,
                               filename=os.path.basename(filename), alerta=alerta, extra_links=extra_links)
    log('Error in editfile with file={}'.format(file), 'error', False)
    return abort(404)


@app.route('/api/uploadfile/<file>', methods=['POST'])
@auto.doc()
def uploadfile(file):
    """
    POST method for VIP files upload & replacement

    :param file: uploaded file_id
    """
    ok_fileid, is_logfile, filename = _get_filepath_from_file_id(file)
    if ok_fileid:
        if is_logfile:
            log('uploading logfile {} [{}]!!!'.format(file, filename), 'error', True)
            return redirect(url_for('index'), code=404)
        f = request.files['file']
        alert = check_uploaded_config_file(file, f, dest_filepath=filename)
        if alert:
            alert = json.dumps(alert)
        return redirect(url_for('editfile', file=file, alerta=alert))
    return abort(500)


#############################
# ENERPI MAILING
#############################
@app.route('/api/email', methods=['GET'])
def send_email():
    text_msg = 'enerPI STATUS from last 24h... '
    alert = 'success'
    mask = os.path.join(STATIC_PATH, "img/generated/tile_enerpi_data_{}_last_24h.svg")
    data_monitor = SENSORS.to_dict()
    for i, k in enumerate(data_monitor['sensors']):
        try:
            data_monitor['sensors'][i]['tile'] = open(mask.format(data_monitor['sensors'][i]['name']), 'r').read()
        except OSError:
            data_monitor['sensors'][i]['tile'] = 'NO TILE'
    try:
        data_monitor['consumption'] = {'tile': open(mask.format('kWh'), 'r').read()}
    except OSError:
        data_monitor['consumption'] = {'tile': 'NO kWh TILE'}
    try:
        data_monitor['ref'] = {'tile': open(mask.format('ref'), 'r').read()}
    except OSError:
        data_monitor['ref'] = {'tile': 'NO REF TILE'}
    html_msg = render_template('email/status_email.html', data_monitor=data_monitor, msg=text_msg)

    msg = Message("Status Info",
                  recipients=[RECIPIENT],
                  sender='"enerPI" <{}>'.format(GMAIL_ACCOUNT),
                  # cc=None,
                  # bcc=GMAIL_ACCOUNT,
                  # attachments=None,
                  reply_to='"enerPI" <{}>'.format(GMAIL_ACCOUNT),
                  # date=None,
                  body=text_msg, html=html_msg)
    # with app.open_resource("static/img/generated/tile_enerpi_data_power_1_last_48h.svg") as fp:
    #     msg.attach("last_48h.svg", "image/svg+xml", fp.read())
    with open(os.path.join(DATA_PATH, INDEX_DATA_CATALOG), 'rb') as fp:
        msg.attach(INDEX_DATA_CATALOG, "text /csv", fp.read())
    log('SENDING STATUS EMAIL', 'debug', True)
    mail.send(msg)
    log('EMAIL SENDED', 'debug', True)
    return redirect(url_for('control', alerta=json.dumps({'alert_type': alert,
                                                          'texto_alerta': 'email content: "{}"'.format(text_msg)})))



# -*- coding: utf-8 -*-
import argparse
import os
from time import time, sleep
# do this before importing pylab or pyplot
import matplotlib
matplotlib.use('Agg')
# noinspection PyPep8
from enerpi.base import (check_resource_files, log, set_logging_conf,
                         STATIC_PATH, SERVER_FILE_LOGGING_RSCGEN, LOGGING_LEVEL, IMG_TILES_BASEPATH)
# noinspection PyPep8
from enerpiplot.enerplot import gen_svg_tiles
# noinspection PyPep8
from enerpi.database import init_catalog


VERBOSE = False
TILES_GENERATION_INTERVAL_LOOP = 180

# Establecemos logging
LOGGING_LEVEL_SERVER = LOGGING_LEVEL
set_logging_conf(SERVER_FILE_LOGGING_RSCGEN, LOGGING_LEVEL_SERVER, with_initial_log=True, verbose=VERBOSE)

basedir_enerweb = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'enerpiweb')
check_resource_files(STATIC_PATH, os.path.join(basedir_enerweb, 'static'), verbose=VERBOSE)
check_resource_files(os.path.join(IMG_TILES_BASEPATH, 'any_image'), verbose=VERBOSE)


###############################
# RESOURCE GENERATOR
###############################
def _rsc_generator(catalog, verbose=False):
    tic, ok = time(), False
    log('**RESOURCE_GENERATOR desde MULE con PID={}!!!'.format(os.getpid()), 'debug', verbose)
    ok = gen_svg_tiles(IMG_TILES_BASEPATH, catalog, last_hours=(72, 48, 24))
    toc = time()
    log('(MULE) TILES generation ok? {}. TOOK {:.3f} s'.format(ok, toc - tic), 'info', verbose)
    return ok, toc - tic


def _loop_rsc_generator(catalog, sleep_time, verbose=False):
    log('**RESOURCE_GENERATOR LOOP desde MULE con PID={}!!!'.format(os.getpid()), 'debug', False)
    counter = 0
    while True:
        gen_ok, took = _rsc_generator(catalog, verbose=verbose)
        counter += 1
        sleep(max(sleep_time - took, 20))
        # uwsgi.async_sleep(max(sleep_time - took, 20))


def main():
    """
    CLI Manager for generate svg plots as resources for the web app
    """
    p = argparse.ArgumentParser(description="EnerPI Web Resource generator")
    p.add_argument('-o', '--one', action='store_true', help='Run only one time.')
    p.add_argument('-t', '--timer', action='store', type=int, metavar='∆T', default=TILES_GENERATION_INTERVAL_LOOP,
                   help='Set periodic timer, in seconds. Default={}.'.format(TILES_GENERATION_INTERVAL_LOOP))
    p.add_argument('-v', '--verbose', action='store_true', help='Verbose mode (logging to stdout)')
    args = p.parse_args()
    # logging.debug('(MULE)-> main con args={}'.format(args))

    # Cargamos catálogo para consulta:
    cat = init_catalog(check_integrity=False)

    if args.one:
        _rsc_generator(cat, verbose=args.verbose)
    else:
        log('(MULE)-> Init loop with timer={} seconds'.format(args.timer), 'debug', False)
        # print('(MULE)-> Init loop with timer={} seconds'.format(args.timer))
        _loop_rsc_generator(cat, args.timer, VERBOSE)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Methods for edit configuration files in ENERPI WebServer, with some validation

    - web_edit_* := from filepath to lines_file + dict_form_template
    - web_post_* := from posted form with changes and original config for saving user config

"""
from collections import OrderedDict
from io import BytesIO
import json
from jsondiff import diff
import os
import textwrap
from enerpi.base import log, get_lines_file, ENCODING, DATA_PATH, CONFIG
from enerpi.iobroadcast import get_codec


ENERPI_CONFIG_FILES = {
    'config': {
        'order': 0,
        'filename': 'config_enerpi.ini',
        'text_button': '<i class="fa fa-check-square-o" aria-hidden="true"></i> ENERPI config',
        'show_switch_comments': True,
    },
    'sensors': {
        'order': 1,
        'filename': 'sensors_enerpi.json',
        'text_button': '<i class="fa fa-wrench" aria-hidden="true"></i> SENSORS',
        'show_switch_comments': False,
    },
    'encryption_key': {
        'order': 2,
        'filename': os.path.join(DATA_PATH, CONFIG.get('BROADCAST', 'KEY_FILE', fallback='.secret_key')),
        'text_button': '<i class="fa fa-user-secret" aria-hidden="true"></i> Secret Key',
        'show_switch_comments': False,
    },
}

TITLE_EDIT_CRYPTOKEY = 'Encryption KEY'
SUBTITLE_EDIT_CRYPTOKEY = 'KEY_FILE'
TITLE_EDIT_JS_SENSORS = 'SENSORS'
SUBTITLE_EDIT_JS_SENSORS = 'SENSORS'


# TODO readonly/disabled attr in protected fields of INI file
def _web_edit_enerpi_config_ini(lines_ini_file):
    """
    * Divide configuration INI file with this structure:
    [section]
    # comment
    ; comment
    VARIABLE = VALUE
    (discarding file header comments)

    Make Ordered dict like:
    ==> [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         (section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         ...]

    :param lines_ini_file: Text lines of INI file
    :return: :tuple of (lines_file_for_webview, dict_file_for_webform):

    """
    def _is_bool_variable(value):
        value = value.lower()
        if (value == 'true') or (value == 'false'):
            return True, value == 'true'
        return False, None

    config_entries = OrderedDict()
    section, comment = None, None
    init = False
    for l in lines_ini_file:
        l = l.replace('\n', '').lstrip().rstrip()
        if l.startswith('['):
            section = l.replace('[', '').replace(']', '')
            init = True
            comment = None
            config_entries[section] = OrderedDict()
        elif l.startswith('#') or l.startswith(';'):
            if init:
                if comment is None:
                    comment = l[1:].lstrip()
                else:
                    comment += ' {}'.format(l[1:].lstrip())
        elif init and (len(l) > 0):
            # Read variable and append (/w comments)
            variable_name, variable_value = l.split('=')
            variable_name = variable_name.lstrip().rstrip()
            variable_value = variable_value.lstrip().rstrip()
            is_bool, bool_value = _is_bool_variable(variable_value)
            if not is_bool:
                try:
                    variable_value = int(variable_value)
                    var_type = 'int'
                except ValueError:
                    try:
                        variable_value = float(variable_value)
                        var_type = 'float'
                    except ValueError:
                        var_type = 'str'
            else:
                var_type = 'bool'
                variable_value = bool_value
            config_entries[section][variable_name] = variable_value, var_type, comment
            comment = None
    return True, config_entries


def _web_edit_enerpi_sensors_json(lines_json):
    """
    * Get JSON sensors config for web user edition:

    Make Ordered dict like:
    ==> [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         (section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         ...]

    :param lines_json: Lines of JSON file
    :return: :tuple of (ok, dict_file_for_webform):

    """
    # TODO JSON EDITOR
    try:
        sensors_json = json.loads('\n'.join(lines_json), encoding=ENCODING)
        t, sub = TITLE_EDIT_JS_SENSORS, SUBTITLE_EDIT_JS_SENSORS
        d_conf = OrderedDict([(t, OrderedDict([(sub, (json.dumps(sensors_json, indent=1), 'text', None))]))])
        return True, d_conf
    except json.decoder.JSONDecodeError as e:
        msg = 'JSONDecodeError reading new SENSORS JSON: {} --> {}'.format(e, lines_json)
        log(msg, 'error')
        return False, {'error': msg}


def _web_edit_enerpi_encryption_key(key_file_lines):
    """
    Return dict for user edition of encryption key
    Make Ordered dict like:
    ==> [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'text', comment=None))])]

    :param key_file_lines: Text lines of encryption key file (1)
    :return: :tuple of (lines_file_for_webview, dict_file_for_webform):

    """
    try:
        assert len(key_file_lines) == 1
        key = key_file_lines[0]
    except AssertionError:
        msg = 'ERROR Reading CRYPTO Key (incorrect # of lines): {}'.format(key_file_lines)
        log(msg, 'error', False)
        return False, {'error': msg}
    try:
        _ = get_codec(key.encode())
    except AssertionError as e:
        msg = 'ASSERT WITH CRYPTO_KEY: {}, KEY="{}"'.format(e, key)
        log(msg, 'error', False)
        return False, {'error': msg}
    t, sub = TITLE_EDIT_CRYPTOKEY, SUBTITLE_EDIT_CRYPTOKEY
    return [key], OrderedDict([(t, OrderedDict([(sub, (key, 'text', None))]))])


def _make_ini_file(dict_config, dest_path=None):
    """
    Makes INI file (and writes it if dest_path is not None) from an OrderedDict
    like the one 'config_dict_for_web_edit' returns.

    * INI file with this structure:
    [section]
    # comment
    ; comment
    VARIABLE = VALUE

    Ordered dict like:
    ==> [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         (section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         ...]

    :param dict_config: :OrderedDict: Content of INI file after 'readlines()'
    :param dest_path: :str: optional path for INI file write.
    :return: :str: raw INI text

    """
    lines = ['# -*- coding: utf-8 -*-']
    for section, entries in dict_config.items():
        lines.append('[{}]'.format(section.upper()))
        for variable_name, (value, var_type, comment) in entries.items():
            if comment:
                [lines.append('# {}'.format(l_wrap)) for l_wrap in textwrap.wrap(comment, 80)]
            lines.append('{} = {}'.format(variable_name, value))
        lines.append('')  # Separador entre secciones
    ini_text = '\n'.join(lines)
    if dest_path is not None:
        with open(dest_path, 'w') as f:
            f.write(ini_text)
    return ini_text


def _web_post_changes_enerpi_config_ini(dict_web_form, lines_config, dict_config, dest_filepath=None):
    """
    Process changes in web editor of INI ENERPI config file

    :param dict_web_form: :OrderedDict: Posted Form values
    :param lines_config:  :list: original file text lines
    :param dict_config:  :OrderedDict: original config dict of dicts (like the one 'web_edit_enerpi_config_ini' returns)
    :param dest_filepath:  :str: (optional) destination filepath for save configuration changes
    :return: :tuple: (:dict: alert message, :list: text lines, :OrderedDict: updated dict_config))

    """
    def _is_changed(value, params, name):
        if (value is None) or (len(value) == 0):
            msg = 'INI Value ERROR: key={}, value="{}", (type={})'.format(name, value, params[1])
            return False, value, msg
        try:
            if params[1] == 'int':
                try:
                    value = int(value)
                except ValueError:
                    value = float(value)
            elif params[1] == 'float':
                value = float(value)
            elif params[1] == 'bool':
                value = (value.lower() == 'true') or (value.lower() == 'on')
            if value != params[0]:
                log('"{}" -> HAY CAMBIO DE {} a {} (type={})'.format(name, params[0], value, params[1]), 'debug', False)
                return True, value, None
            return False, value, None
        except ValueError as e:
            msg = 'ValueError reading INI: key={}, value={}, (type={}). Error: {}'.format(name, value, params[1], e)
            return False, value, msg

    dict_config_updated = dict_config.copy()
    dict_web_form = dict_web_form.copy()
    vars_updated = []
    for section, entries in dict_config_updated.items():
        for variable_name, variable_params in entries.items():
            if variable_name in dict_web_form:
                new_v = dict_web_form.pop(variable_name)
                changed, new_value, error = _is_changed(new_v, variable_params, variable_name)
                if error is not None:
                    alerta = {'alert_type': 'danger', 'texto_alerta': error}
                    return alerta, lines_config, dict_config
                if changed:
                    vars_updated.append((variable_name, variable_params[0], new_value))
                    params_var = list(dict_config_updated[section][variable_name])
                    params_var[0] = new_value
                    dict_config_updated[section][variable_name] = tuple(params_var)
            elif (variable_params[1] == 'bool') and variable_params[0]:  # Bool en off en el form y True en config
                vars_updated.append((variable_name, variable_params[0], False))
                params_var = list(dict_config_updated[section][variable_name])
                params_var[0] = False
                log('"{}" -> HAY CHECKBOX CH DE {} a {} (type={})'
                    .format(variable_name, variable_params[0], False, variable_params[1]), 'debug', False)
                dict_config_updated[section][variable_name] = tuple(params_var)
    alerta = None
    if len(vars_updated) > 0:
        str_cambios = ('Configuration changes in:<br>{}<br> New config SAVED!'
                       .format('<br>'.join(['- <strong>"{}"</strong> (before={}) -> <strong>"{}"</strong>'
                                           .format(name, ant, new) for name, ant, new in vars_updated])))
        alerta = {'alert_type': 'warning', 'texto_alerta': str_cambios}
        log(str_cambios, 'debug', False)
        lines_config = _make_ini_file(dict_config_updated, dest_path=dest_filepath).splitlines()
    return alerta, lines_config, dict_config_updated


def _web_post_changes_enerpi_sensors_json(dict_web_form, lines_config, dict_config, dest_filepath=None):
    """
    Process changes in web editor of ENERPI SENSORS JSON File

    :param dict_web_form: :OrderedDict: Posted Form with new JSON data
    :param lines_config: :list: original file text lines (1) with original JSON
    :param dict_config: :OrderedDict: original config dict of dicts
                            (like the one 'web_edit_enerpi_sensors_json' returns)
    :param dest_filepath: :str: (optional) destination filepath for save configuration changes
    :return: :tuple: (:dict: alert message, :list: text lines, :OrderedDict: updated dict_config))

    """
    alerta = None
    t, sub = TITLE_EDIT_JS_SENSORS, SUBTITLE_EDIT_JS_SENSORS
    try:
        ant_json = json.loads(dict_config[t][sub][0], encoding=ENCODING)
        new_json = json.loads(dict_web_form[sub], encoding=ENCODING)
        diff_json = diff(ant_json, new_json)
        if diff_json:
            if dest_filepath is not None:
                log('New JSON Sensors config ("{}")\nwill be saved in "{}"'
                    .format(new_json, dest_filepath), 'warning', False)
                with open(dest_filepath, 'w') as f:
                    json.dump(new_json, f, indent=1)
            str_cambios = ('Configuration changes in ENERPI SENSORS:<br>{}<br> New config SAVED!'
                           .format('JSON DIFF- <strong>"{}"</strong>'.format(diff_json)))
            log(str_cambios, 'debug', False)
            alerta = {'alert_type': 'warning', 'texto_alerta': str_cambios}
            if dest_filepath is not None:
                lines_config = get_lines_file(dest_filepath)
                ok, dict_config = _web_edit_enerpi_sensors_json(lines_config)
                if not ok:
                    alerta['alert_type'] = 'error'
                    alerta['texto_alerta'] += dict_config['error']
                    alerta['texto_alerta'] += '\nNEW JSON SENSORS FILE NOT VALID!! FIX IT, PLEASE'
    except json.decoder.JSONDecodeError:
        msg_err = ('JSONDecodeError in web_post_changes_enerpi_sensors_json: {}'.format(dict_config[t][sub]))
        alerta = {'alert_type': 'error', 'texto_alerta': msg_err}
    return alerta, lines_config, dict_config


def _web_post_changes_enerpi_encryption_key(dict_web_form, lines_keyfile, dict_key_file, dest_filepath=None):
    """
    Process changes in web editor of ENERPI BROADCAST ENCRYPTION KEY
    # Secret Key edition!! NOT SAFE AT ALL!! (Debugging purposes)

    :param dict_web_form: :OrderedDict: Posted Form with new crypto key
    :param lines_keyfile: :list: original file text lines (1) with original cryto key
    :param dict_key_file: :OrderedDict: original config dict of dicts
                            (like the one 'web_edit_enerpi_encryption_key' returns)
    :param dest_filepath: :str: (optional) destination filepath for save configuration changes
    :return: :tuple: (:dict: alert message, :list: text lines, :OrderedDict: updated dict_config))

    """
    t, sub = TITLE_EDIT_CRYPTOKEY, SUBTITLE_EDIT_CRYPTOKEY
    ant_key = dict_key_file[t][sub][0]
    new_key = dict_web_form[sub]
    alerta = {}
    if new_key != ant_key:
        try:  # Validate new Fernet Key
            _ = get_codec(new_key.encode())
            if len(new_key) > 10:
                if dest_filepath is not None:
                    log('The new KEY ("{}")\nwill be saved in "{}"\n'.format(new_key, dest_filepath), 'warning', False)
                    with open(dest_filepath, 'wb') as f:
                        f.write(new_key.encode())
                str_cambios = ('Configuration changes in encryption key:<br>{}<br> New config SAVED!'
                               .format('- <strong>"{}"</strong> (before={}) -> <strong>"{}"</strong>'
                                       .format(SUBTITLE_EDIT_CRYPTOKEY, ant_key, new_key)))
                log(str_cambios, 'debug', False)
                alerta = {'alert_type': 'warning', 'texto_alerta': str_cambios}
                lines_keyfile = [new_key]
                dict_key_file = OrderedDict([(t, OrderedDict([(sub, (new_key, 'text', None))]))])
        except AssertionError as e:
            alerta = {'alert_type': 'danger', 'texto_alerta': 'Not a valid KEY: {}. New KEY was: {}'.format(e, new_key)}
    return alerta, lines_keyfile, dict_key_file


def check_uploaded_config_file(file_id, f_obj, dest_filepath=None):
    """Check uploaded configuration file and save it if correct. Return dict for web alert

    :param file_id: config file id
    :param f_obj: :werkzeug.datastructures.FileStorage: uploaded file object
    :param dest_filepath: optional destination path for uploaded file

    :return: (web alert)
    """
    file_name, file_extension = os.path.splitext(f_obj.filename)
    fmem = BytesIO()
    f_obj.save(fmem)
    fmem.seek(0)
    content = fmem.read().decode()
    lines = content.splitlines()
    if len(content) < 20:
        msg = 'Incorrect uploaded file (too short!): {}'.format(content)
        return {'alert_type': 'danger', 'texto_alerta': msg}
    if (file_id == 'config') and (file_extension == '.ini') and (f_obj.mimetype == 'application/octet-stream'):
        ok_file, error = _web_edit_enerpi_config_ini(lines)
    elif (file_id == 'sensors') and (file_extension == '.json') and (f_obj.mimetype == 'application/json'):
        ok_file, error = _web_edit_enerpi_sensors_json(lines)
    elif (file_id == 'encryption_key') and (f_obj.mimetype == 'application/octet-stream'):
        ok_file, error = _web_edit_enerpi_encryption_key(lines)
    else:
        return {'alert_type': 'danger', 'texto_alerta': 'FileId not recognized: {}'.format(file_id)}
    if not ok_file:
        return {'alert_type': 'danger', 'texto_alerta': error['error']}
    if dest_filepath is not None:
        with open(dest_filepath, 'w') as f:
            f.write(content)
    return {'alert_type': 'success',
            'texto_alerta': 'Uploaded file "{}" as replacement in "{}"!'.format(f_obj.filename, dest_filepath)}


def web_config_edition_data(file_id, filepath, d_edition_form=None):
    """
    Method for validate and save changes on configuration files.
    It returns data for jinja2 templates:
        alert message, Data Get JSON sensors config for web user edition:

    :param file_id: Id of config file
    :param filepath: filepath for read/save configuration changes
    :param d_edition_form: request.form dict for config POST requests as:
        [(section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         (section_name,
            OrderedDict([(VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment)),
                         ...
                         (VARIABLE, (VALUE, 'int|float|bool|text', comment))])),
         ...]
    :return: (:dict: alert message, :list: text lines, :OrderedDict: dict_config for jinja2 templates))

    """
    if file_id == 'config':
        f_edit_file = _web_edit_enerpi_config_ini
        f_post_file = _web_post_changes_enerpi_config_ini
    elif file_id == 'sensors':
        f_edit_file = _web_edit_enerpi_sensors_json
        f_post_file = _web_post_changes_enerpi_sensors_json
    else:
        assert file_id == 'encryption_key'
        f_edit_file = _web_edit_enerpi_encryption_key
        f_post_file = _web_post_changes_enerpi_encryption_key
    alerta = None
    lines_config = get_lines_file(filepath)
    ok_lines, config_data = f_edit_file(lines_config)
    if not ok_lines:
        alerta = {'alert_type': 'danger', 'texto_alerta': config_data['error']}
        log('NOT OK in web_config_edition_data: {}'.format(web_config_edition_data), 'warning', False)
    elif d_edition_form is not None:
        alerta, lines_config, config_data = f_post_file(d_edition_form, lines_config, config_data, filepath)
    return alerta, lines_config, config_data



# -*- coding: utf-8 -*-
from codecs import open
import os
from setuptools import setup, find_packages
from enerpi import __version__ as version, BASE_PATH


# Failed building wheel for cryptography, cffi
# Si error de cffi en install:
# sudo apt-get install build-essential libssl-dev libffi-dev python-dev

packages = find_packages(exclude=['docs', '*tests*', 'notebooks'])
with open(os.path.join(BASE_PATH, '..', 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='enerpi',
    version=version,
    description='AC Current Meter for Raspberry PI with GPIOZERO and MCP3008',
    long_description='\n' + long_description,
    keywords='enerpi current gpiozero raspberry analog',
    author='Eugenio Panadero',
    author_email='azogue.lab@gmail.com',
    url='https://github.com/azogue/enerpi',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Home Automation',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Natural Language :: Spanish',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Operating System :: Unix'],
    packages=packages,
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'enerpi': ['config/default_config_enerpi.ini', 'config/.enerpi_data_path'],
        'enerpiweb': ['templates/*', 'static/css/*', 'static/img/icons/*', 'static/js/*'],
    },
    install_requires=['numpy>=1.11.2', 'pandas>=0.19.0', 'pytz>=2016.7',
                      'flask>=0.11.1', 'flask-autodoc>=0.1.2', 'wtforms>=2.1',  'cryptography>=1.5.2',
                      'gpiozero>=1.3.1', 'RPi.GPIO>=0.6.3',
                      'matplotlib>=1.5.3', 'bokeh>=0.12.3', 'cairosvg>=1.0.22', 'termcolor>=1.1.0',
                      'python-crontab', 'jsondiff>=1.0.0'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    entry_points={
        'console_scripts': [
            'enerpi = enerpi.__main__:enerpi_main_cli',
            'pitemps = enerpi.pitemps.__main__:main',
            'enerpi-daemon = enerpi.enerdaemon:enerpi_daemon',
            'enerpi-rscgen= enerpiplot.mule_rscgen:main',
            'enerpiweb = enerpiweb.__main__:main'
        ]
    },
    # Test configuration:
    # setup_requires=['pytest-runner'],
    tests_require=['pytest>=3.0.0'],
    # test_suite='nose.collector',
)



# -*- coding: utf-8 -*-
from enerpi.pitemps import get_cpu_temp, get_gpu_temp


def main():
    """
    Prints RPI temperatures on sys.stdout.

    """
    print("CPU temp: ", str(get_cpu_temp()))
    print("GPU temp: ", str(get_gpu_temp()))


if __name__ == '__main__':
    main()
